import argparse
import itertools
import os
import random
import re
import string
import threading
import time
from typing import List

import keras
import nltk
import numpy as np
import tensorflow as tf
import unidecode
from gensim.models.wrappers import FastText as FastTextWrapper
from keras import backend as K
from tensorflow import keras
from rb.parser.spacy_parser import SpacyParser
from rb.core.lang import Lang
from rb.utils.rblogger import Logger
nltk.download('punkt')

logger = Logger.get_logger()


tf.autograph.set_verbosity(0)

class Diacritics:

	def __init__(self):
		# set verbosity level */
		tf.logging.set_verbosity(tf.logging.FATAL)
		self.model_embeddings = None
		self.dict_lock = threading.Lock()
		self.dict_avg_words = {}

		# the weights of the model will be saved in a folder per each epoch
		self.fast_text = "fastText/wiki.ro"

		self.window_sentence = 15 
		self.window_character = 6 # 2 * x + 1

		self.character_embedding_size = 28
		self.tag_embedding_size = 5
		self.dep_embedding_size = 5
		self.word_embedding_size = 300


		self.characters_cell_size = 64
		self.sentence_cell_size = 300
		self.neurons_dense_layer_after_merge = 512
		self.batch_size = 128
		self.limit_backtracking_characters = 10

		self.CPUS = 1
		self.size_prefetch_buffer = 10
		self.max_unicode_allowed = 770
		self.replace_character = 255
		self.padding_character = 0

		self.train_files = "corpus/train/"
		self.test_files = "corpus/test/"
		self.valid_files = "corpus/validation/"

		self.samples_number = {
			'full_train': 380968863,
			'full_test': 131424533,
			'full_valid': 131861863,
			'par_train': 84321880,
			'par_test': 29407143,
			'par_valid': 28882058,
		}

		self.map_no_diac = {
			'ă': 'a',
			'â': 'a',
			'Â': 'A',
			'Ă': 'A',
			'ț': 't',
			'Ț': 'T',
			'ș': 's',
			'Ș': 'S',
			'î': 'i',
			'Î': 'I',
			"ş": "ș",
			"Ş": "Ș",
			"ţ": "ț",
			"Ţ": "Ț",
		}
		self.all_in = {
			'ă','â','a',
			'Ă','Â','A',	
			'ț',
			'Ț',
			'ș',
			'Ș',
			'î',
			'Î',
			'i', 
			't',
			's',
			"ş",
			"Ş",
			"ţ",
			"Ţ",
			'S',
			'T',
			'I'
		}

		self.map_correct_diac = {
			"ş": "ș",
			"Ş": "Ș",
			"ţ": "ț",
			"Ţ": "Ț",
		}
		self.map_char_to_possible_chars = {
			'a': ['ă', 'â', 'a'], 
			'i': ['î', 'i'],
			's': ['ș', 's'], 
			't': ['ț', 't']
		}
		self.map_substitute_chars = {'"': '`'}
		self.characters_in_interest = {'a', 'i', 's', 't'}
		self.to_lower = {}
		self.rom_spacy = SpacyParser.get_instance().get_model(Lang.RO)
		
		self.create_lower_mapping()
		self.parse_args()

	def get_word_interval_mistake(self, s, index):
		nn = len(s)
		ii = index
		while ii < nn and s[ii] != ' ': ii += 1
		dr = ii - 1
		ii = index
		while ii >= 0 and s[ii] != ' ': ii -= 1 
		st = ii + 1
		return st, dr

	# get case of the highest probability (returned by softmax)
	def get_case(self, p):
		case = 0
		for i in range(len(p)):
			if p[i] > p[case]:
				case = i
		return case

	def create_lower_mapping(self):
		for c in string.ascii_uppercase:
			self.to_lower[c] = c.lower()
		self.to_lower['Ș'] = 'ș'
		self.to_lower['Â'] = 'â'
		self.to_lower['Ă'] = 'ă'
		self.to_lower['Ț'] = 'ț'
		self.to_lower['Î'] = 'î'

	# case:
	#	0 -> ă
	#	1 -> â, î
	#	2 -> unmodified
	# 	3 -> ș, ț
	def get_label(self, i, clean_text_utf, original_text_utf):
		case = 2
		if clean_text_utf[i] == 'a':
			if original_text_utf[i] == 'ă':
				case = 0
			elif original_text_utf[i] == 'â':
				case = 1
			elif original_text_utf[i] == 'a':
				case = 2
		elif clean_text_utf[i] == 'i':
			if original_text_utf[i] == 'î':
				case = 1
			elif original_text_utf[i] == 'i':
				case = 2
		elif clean_text_utf[i] == 't':
			if original_text_utf[i] == 'ț':
				case = 3
			elif original_text_utf[i] == 't':
				case = 2
		elif clean_text_utf[i] == 's':
			if original_text_utf[i] == 'ș':
				if self.args.nr_classes == 5:
					case = 4
				else:
					case = 3
			elif original_text_utf[i] == 's':
				case = 2

		label = np.float32([0] * self.args.nr_classes)
		label[case] = np.float32(1.0)
		return label

	def bkt_all_words(self, index: int, clean_word: str, current_word: List) -> List:

		if index == len(clean_word):
			word = "".join(current_word)
			if self.args.use_dummy_word_embeddings == True:
				return [word]
			else:
				if word in self.model_embeddings.wv.vocab:
					return [word]
				else:
					return []
		else:
			L = []
			c = clean_word[index]
			if c in self.map_char_to_possible_chars:
				for ch in self.map_char_to_possible_chars[c]:
					current_word[index] = ch
					L += self.bkt_all_words(index + 1, clean_word, current_word)
			else:
				current_word[index] = c
				L += self.bkt_all_words(index + 1, clean_word, current_word)
			return L

	def get_avg_possible_word(self, clean_word):

		with self.dict_lock:
			if clean_word in self.dict_avg_words:
				return self.dict_avg_words[clean_word]
			
		count_diacritics_chars = 0
		for c in clean_word:
			if c in self.map_char_to_possible_chars:
				count_diacritics_chars += 1
		
		if count_diacritics_chars > self.limit_backtracking_characters:
			if self.args.use_dummy_word_embeddings == False:
				return np.float32(self.model_embeddings.wv[clean_word])
			else:
				return np.float32([0] * self.word_embedding_size)

		all_words = self.bkt_all_words(0, clean_word, ['a'] * len(clean_word))
		
		if len(all_words) > 0:
			if self.args.use_dummy_word_embeddings == False:
				return np.mean([np.float32(self.model_embeddings.wv[word]) for word in all_words], axis=0)
			else:
				return np.float32([0] * self.word_embedding_size)
		else:
			try:
				if self.args.use_dummy_word_embeddings == False:
					return np.float32(self.model_embeddings.wv[clean_word])
				else:
					return np.float32([0] * self.word_embedding_size)
			except:
				return np.float32([0] * self.word_embedding_size)

	def get_embeddings_sentence(self, clean_tokens_sentence, index_token):
		embeddings_sentence = []

		for i in range(index_token - self.window_sentence, index_token + self.window_sentence + 1):
			if i >= 0 and i < len(clean_tokens_sentence):
				token = clean_tokens_sentence[i]
				token_embedding = self.get_avg_possible_word(token)
				embeddings_sentence.append(token_embedding)
				with self.dict_lock:
					self.dict_avg_words[token] = token_embedding
			else:
				embeddings_sentence.append(np.float32([0] * self.word_embedding_size))
		return np.array(embeddings_sentence)

	# return an tuple input (window_char, embedding_token, embedding_sentence)
	def get_input_example(self, clean_text_utf, index_text, clean_tokens, index_sent, \
					index_last_sent, index_token):
		# window with characters
		w = []
		for j in range(index_text - self.window_character, index_text + self.window_character + 1):
			if j < 0 or j >= len(clean_text_utf):
				v1 = self.padding_character
			elif ord(clean_text_utf[j]) > self.max_unicode_allowed:
				v1 = self.replace_character
			else:
				v1 = ord(clean_text_utf[j])
			w.append(v1)
		# token 
		token = clean_tokens[index_sent][index_token]
		token_embedding = self.get_avg_possible_word(token)
		with self.dict_lock:
			self.dict_avg_words[token] = token_embedding

		# sentence 
		# if is the same sentence don't recompute it
		if index_last_sent is None or index_sent != index_last_sent:
			sentence_embedding = self.get_embeddings_sentence(clean_tokens[index_sent], index_token)
		else:
			sentence_embedding = None
		return (np.int32(w), token_embedding, sentence_embedding)

	def replace_char(self, c):

		if c in self.map_correct_diac:
			c = self.map_correct_diac[c]

		if c in self.to_lower:
			c = self.to_lower[c]

		if c in self.map_no_diac:
			c = self.map_no_diac[c]

		if ord(c) > 255:
			return chr(self.replace_character)
		elif c in self.map_substitute_chars:
			return self.map_substitute_chars[c]
		else:
			return c

	def replace_char_original(self, c):
		if c in self.map_correct_diac:
			c = self.map_correct_diac[c]

		if c in self.to_lower:
			c = self.to_lower[c]
		return c

	def count_chars_in_interest(self, s):
		cnt_chars = 0
		chars_in_int = []
		for c in s:
			character_c = chr(c)
			if character_c in self.characters_in_interest:
				cnt_chars += 1
				chars_in_int.append(character_c)
		return cnt_chars, chars_in_int

	def get_word_around_index(self, original_text_utf, index_text):
		s = []
		i = index_text
		while i >= 0 and original_text_utf[i].isalpha():
			i -= 1
		i += 1
		while i < len(original_text_utf) and original_text_utf[i].isalpha():
			s.append(original_text_utf[i])
			i += 1
		return "".join(s)

	def create_examples(self, original_text, is_test_dataset):
		drop_example = False

		try:
			original_text_utf = original_text.decode('utf-8')
			first_clean_text_utf = "".join([self.replace_char_original(c) for c in original_text_utf])
			# replace some strange characters which are modified by tokenization
			clean_text_utf = "".join([self.replace_char(c) for c in first_clean_text_utf])
			clean_sentences = nltk.sent_tokenize(clean_text_utf)
			clean_tokens = []
			clean_tokens_tag = []
			clean_tokens_dep = []

			# construct tokens
			for i in range(len(clean_sentences)):
				tokens = self.rom_spacy(clean_sentences[i])

				clean_tokens_sent = [token.text for token in tokens]
				tag_tokens_sent = [int(token.tag) for token in tokens]
				dep_tokens_sent = [int(token.dep) for token in tokens]
				
				#clean_tokens_sent = nltk.word_tokenize(clean_sentences[i])
				clean_tokens.append(clean_tokens_sent)
				clean_tokens_tag.append(tag_tokens_sent)
				clean_tokens_dep.append(dep_tokens_sent)

			index_text = 0 # current position in text
			index_sent = 0 # current sentence
			index_token = 0 # current token
			index_last_sent = None # last sentence computed

			# input and output lists
			clean_words = []
			original_words = []
			self.window_characters = []
			word_embeddings = []
			sentence_embeddings = []
			tag_tokens = []
			dep_tokens = []
			labels = []

			while index_sent < len(clean_tokens):
				clean_token = clean_tokens[index_sent][index_token]
				tag_token = clean_tokens_tag[index_sent][index_token]
				dep_token = clean_tokens_dep[index_sent][index_token]

				original_word = self.get_word_around_index(original_text_utf, index_text)
				i = 0
				# construct all inputs from the current token (letter(a, i, t, s) -> input)
				while i < len(clean_token):

					if clean_text_utf[index_text] in self.characters_in_interest:

						label = self.get_label(index_text, clean_text_utf, original_text_utf)
						#print(original_text_utf[index_text], label)
						win_char, word_emb, sent_emb = self.get_input_example(clean_text_utf, \
								index_text, clean_tokens, index_sent, index_last_sent, index_token)
						index_last_sent = index_sent
						if is_test_dataset == True:
							clean_words.append(clean_token)
							original_words.append(original_word)

						self.window_characters.append(win_char)
						word_embeddings.append(word_emb)
						tag_tokens.append(np.float32([tag_token]))
						dep_tokens.append(np.float32([dep_token]))
						# sentence already computed
						if sent_emb is None:
							sentence_embeddings.append(sentence_embeddings[-1])
						else:
							sentence_embeddings.append(sent_emb)					
						labels.append(label)
						#print(clean_text_utf[index_text], original_text_utf[index_text], label)

					if clean_text_utf[index_text] == clean_token[i]:
						index_text += 1
						i += 1
					else: # discard char in text
						index_text += 1

				if index_token == len(clean_tokens[index_sent]) - 1:
					index_token = 0
					index_sent += 1
				else:
					index_token += 1
		except Exception as e:
			drop_example = True
				
		# dummy values for empty sentence
		if len(self.window_characters) == 0 or drop_example == True:
			clean_words = ['a']
			self.window_characters = [np.int32([0] * (self.window_character * 2 + 1))]
			word_embeddings = [np.float32([0] * self.word_embedding_size)]
			sentence_embeddings = [np.array(\
				[np.float32([0] * self.word_embedding_size)] * (self.window_sentence * 2 + 1))]
			tag_tokens = [np.float32([0])]
			dep_tokens = [np.float32([0])]
			lab = np.float32([0] * self.args.nr_classes)
			lab[2] = np.float32(1.0)
			labels = [lab]
		if is_test_dataset == False:
			return (self.window_characters, word_embeddings, sentence_embeddings, tag_tokens, dep_tokens, labels)
		else:
			return (clean_words, self.window_characters, word_embeddings, sentence_embeddings, tag_tokens, dep_tokens, labels, original_words)

	def filter_null_strings(self, s):
		if len(s) == 0:
			return np.array([False])
		return np.array([True])

	def flat_map_f(self, a, b):
		return tf.data.Dataset.from_tensor_slices((a, b))

	def get_dataset(self, dpath, sess, is_test_dataset=False, restore=False, batch_1=False):

		if restore == False:
			input_files = tf.gfile.ListDirectory(dpath)
			for i in range(len(input_files)):			
				if self.args.corpus_rowiki == False and input_files[i].count('rowiki') > 0:
					input_files.remove(input_files[i])
					break
			
			for i in range(len(input_files)):
				input_files[i] = dpath + input_files[i]
		else:
			input_files = [dpath]

		dataset = tf.data.TextLineDataset(input_files)
		dataset = dataset.filter(lambda x:
		(tf.py_func(self.filter_null_strings, [x], tf.bool, stateful=False))[0])

		if is_test_dataset == True:
			datatype_returned = (tf.string, tf.int32, tf.float32, tf.float32, tf.float32, tf.float32,\
				tf.float32, tf.string)
		else:
			datatype_returned = (tf.int32, tf.float32, tf.float32, tf.float32, tf.float32,\
				tf.float32)

		dataset = dataset.map(lambda x:\
			tf.py_func(self.create_examples,\
						(x, is_test_dataset,),\
						datatype_returned,\
						stateful=False),\
			num_parallel_calls=self.CPUS)
		# map input - output tuple
		if is_test_dataset == True:
			dataset = dataset.map(lambda x1, x2, x3, x4, x5, x6, y1, y2:\
									((x1, x2, x3, x4, x5, x6), (y1, y2)),\
									num_parallel_calls=self.CPUS)
		else:
			dataset = dataset.map(lambda x1, x2, x3, x4, x5, y:\
									((x1, x2, x3, x4, x5), y),\
									num_parallel_calls=self.CPUS)
		dataset = dataset.flat_map(self.flat_map_f)
		
		# do not shuffle or batch test dataset
		if is_test_dataset == True:
			if batch_1 == False:
				dataset = dataset.batch(self.args.batch_size_restore)
			else:
				dataset = dataset.batch(1)
		else:
			dataset = dataset.shuffle(self.args.buffer_size_shuffle)
			dataset = dataset.batch(self.batch_size)
			dataset = dataset.prefetch(self.size_prefetch_buffer)

		return dataset

	def get_charr(self, simple_c, case):

		# 0 : ă
		# 1 : â, î
		# 2 : a, i, t, s
		# 3 : ț, ș
		if simple_c == 'a':
			if case == 0: # ă
				return 'ă'
			elif case == 1: # 
				return 'â'
			elif case == 2:
				return 'a'
		elif simple_c == 'i':
			if case == 1:
				return 'î'
			elif case == 2:
				return 'i'
		elif simple_c == 't':
			if case == 3:
				return 'ț'
			elif case == 2:
				return 't'
		elif simple_c == 's':
			if case == 3 and self.args.nr_classes == 4:
				return 'ș'
			elif case == 4 and self.args.nr_classes == 5:
				return 'ș'
			elif case == 2:
				return 's'

		logger.warning('Diacritics model may not be run with the proepr number of classes.')
		return 'a'

	def compute_prediction(self, correct_case, predicted_case, simple_c, precision_chars, recall_chars):
		correct_char = self.get_charr(simple_c, correct_case)
		pred_char = self.get_charr(simple_c, predicted_case)

		if pred_char == correct_char:
			# correct results
			recall_chars[correct_char][0] += 1
			precision_chars[correct_char][0] += 1
		precision_chars[pred_char][1] += 1
		recall_chars[correct_char][1] += 1

	def get_next_possible(self, all_txt, index_full_text):
		while True:
			index_full_text += 1
			if all_txt[index_full_text] in self.all_in:
				return index_full_text

	# restore char if you have the word
	def restore_char(self, pred_char, original_char, original_word):
		
		# don't modify char if already there (has diacritics)
		if original_char in self.map_no_diac:
			return original_char
		# don't modify the word if is only uppercase letters
		elif original_word.isupper():
			return original_char
		elif original_char.isupper():
			return pred_char.upper()
		return pred_char

	# restore char without word
	def restore_char_simple(self, pred_char, original_char):
		# don't modify char if already there
		if original_char in self.map_no_diac:
			return original_char
		elif original_char.isupper():
			return pred_char.upper()
		return pred_char

	# restoration expects a file
	def restore_file(self, sess, model, file_path, nr_batches, batch_1=False):

		# print('restoring file {}'.format(file_path))

		all_txt = []
		with open(file_path, 'r', encoding='utf-8') as f:
			for line in f:
				all_txt.append(list(line))
		all_txt = list(itertools.chain(*all_txt))

		if batch_1 == True:
			self.batch_sizee = 1
		else:
			self.batch_sizee = self.args.batch_size_restore

		dt_test = self.get_dataset(file_path, sess, True, True, batch_1=batch_1)
		iterator_test = dt_test.make_initializable_iterator()
		
		sess.run(iterator_test.initializer)
		test_inp_pred, _ = iterator_test.get_next()
		_, test_char_window_pred, test_words_pred, test_sentence_pred, _, _ = test_inp_pred
		test_char_window_pred = tf.reshape(test_char_window_pred, [self.batch_sizee, self.window_character * 2 + 1])
		test_words_pred = tf.reshape(test_words_pred, [self.batch_sizee, self.word_embedding_size])
		test_sentence_pred = tf.reshape(test_sentence_pred, [self.batch_sizee, self.window_sentence * 2 + 1, self.word_embedding_size])

		input_list = self.get_input_list(test_char_window_pred, test_words_pred, test_sentence_pred, None, None)
		predictions = model.predict(x=input_list,
					verbose=1,
					steps=nr_batches)
		sess.run(iterator_test.initializer)
		index_full_text = -1

		for current_prediction in range(self.batch_sizee * nr_batches):
			pred_vector = predictions[current_prediction]
			predicted_case = self.get_case(pred_vector)
			index_full_text = self.get_next_possible(all_txt, index_full_text)
			c = self.replace_char_original(all_txt[index_full_text])
			c = self.replace_char(c)
			pred_char = self.get_charr(c, predicted_case)
			all_txt[index_full_text] = self.restore_char_simple(pred_char, all_txt[index_full_text])
		res = "".join(all_txt)
		return res

	# restoration is done in batches
	# the text is splitted into 2 files - one for the bog batches and with batches == 1
	def restore_diacritics(self, sess, model):

		logger.info('Starting restoring diacritics')
		path1 = "".join([random.choice(string.ascii_uppercase + string.digits) for _ in range(10)]) +  '_big_batches.txt'
		path2 = "".join([random.choice(string.ascii_uppercase + string.digits) for _ in range(10)]) + '_small_batches.txt'
		out_file = self.args.output_file_restore

		txt_file = self.args.restore_diacritics
		all_txt = []
		
		with open(txt_file, 'r', encoding='utf-8') as f:
			for line in f:
				all_txt.append(list(line))
		# print('input text read: {} lines'.format(len(all_txt)))
		all_txt = list(itertools.chain(*all_txt))
		all_txt_chars = 0

		# count nr of predictions (all possible)
		for c in all_txt:
			if c in self.all_in:
				all_txt_chars += 1
		nr_batches = all_txt_chars // self.args.batch_size_restore
		partial_txt_chars = 0
		# print('nr batches: {}'.format(nr_batches))

		if all_txt_chars >= self.args.batch_size_restore:

			for i, c in enumerate(all_txt):
				if c in self.all_in:
					partial_txt_chars += 1
					if partial_txt_chars == nr_batches * self.args.batch_size_restore:
						txt_big_batches = all_txt[:i+1]
						txt_small_batches = all_txt[i+1:]
						break
			with open(path1, 'w', encoding='utf-8') as f:
				f.write("".join(txt_big_batches))
				
			with open(path2, 'w', encoding='utf-8') as f:
				f.write("".join(txt_small_batches))
			
			s1 = self.restore_file(sess, model, path1, nr_batches, batch_1=False)
			s2 = self.restore_file(sess, model, path2, all_txt_chars - partial_txt_chars, batch_1=True)
			os.remove(path1)
			os.remove(path2)

			with open(out_file, 'w', encoding='utf-8') as f:
				f.write(s1 + s2)
		else:
			s = self.restore_file(sess, model, txt_file, all_txt_chars, batch_1=True)
			with open(out_file, 'w', encoding='utf-8') as f:
				f.write(s)
		logger.info('Successfully restored diacritics')

	def compute_test_accuracy(self, sess, model):
		dt_test = self.get_dataset(self.test_files, sess, True)
		iterator_test = dt_test.make_initializable_iterator()

		sess.run(iterator_test.initializer)
		nr_test_batches = self.args.number_samples_test

		test_inp_pred, _ = iterator_test.get_next()
		test_string_word_pred, test_char_window_pred, test_words_pred, test_sentence_pred, _, _ = test_inp_pred
		predictions = model.predict(x=[test_char_window_pred, test_words_pred, test_sentence_pred],
					verbose=1,
					steps=nr_test_batches)

		current_test_batch = 0
		sess.run(iterator_test.initializer)
		test_next_element = iterator_test.get_next()
		prediction_index = 0
		total_words = 0
		correct_predicted_words = 0
		correct_predicted_chars = 0
		wrong_predicatd_chars = 0
		precision_chars = {'ă': [0, 0], 'â': [0, 0], 'a': [0, 0], 'i': [0, 0], 'î': [0, 0], 's': [0, 0],\
							'ș': [0, 0], 't': [0, 0], 'ț': [0, 0]}
		recall_chars = {'ă': [0, 0], 'â': [0, 0], 'a': [0, 0], 'i': [0, 0], 'î': [0, 0], 's': [0, 0],\
							'ș': [0, 0], 't': [0, 0], 'ț': [0, 0]}

		wrong_restoration_words  = {}
		correct_restoration_words = {}
		acc_restoration_word = {}
		all_words= set()

		while True:
			try:
				test_inp, test_out = sess.run(test_next_element)
				test_string_word, _, _, _ = test_inp
				current_test_batch += 1
				index_batch = 0
				if current_test_batch % 100 == 0:
					print('batch {} out of {}'.format(current_test_batch, nr_test_batches))
				while index_batch < len(test_string_word):
					# skip last word no matter what
					word = test_string_word[index_batch]
					all_words.add(word)
					nr_chars_in_word, chars_in_int = self.count_chars_in_interest(word)
					if nr_chars_in_word > len(test_string_word) - index_batch:
						prediction_index += len(test_string_word) - index_batch
						break

					correct_prediction_word = True
					for i in range(nr_chars_in_word):
						pred_vector = predictions[prediction_index]
						predicted_case = self.get_case(pred_vector)
						correct_case = self.get_case(test_out[index_batch][0])
						index_batch += 1
						prediction_index += 1

						if predicted_case != correct_case:
							correct_prediction_word = False
							wrong_predicatd_chars += 1
						else:
							correct_predicted_chars += 1
						self.compute_prediction(correct_case, predicted_case, chars_in_int[i], precision_chars, recall_chars)

					total_words += 1

					if correct_prediction_word == True:
						correct_predicted_words += 1
						if word in correct_restoration_words:
							correct_restoration_words[word] += 1
						else:
							correct_restoration_words[word] = 1
					else:
						if word in wrong_restoration_words:
							wrong_restoration_words[word] += 1
						else:
							wrong_restoration_words[word] = 1
						
				if current_test_batch == nr_test_batches:
					break
			except tf.errors.OutOfRangeError:
				break

		index_word = 0
		print('highest missed words: ')
		for key, value in sorted(wrong_restoration_words.items(), key=lambda x: x[1], reverse=True):
			correct = 0
			if key in correct_restoration_words:
				correct = correct_restoration_words[key]
			print("word '"  + key.decode('utf-8') + "' wrong: " +  str(value) + \
				' correct: ' + str(correct) + ' accuracy: ' + str(1.0 * correct / (value + correct)))
			index_word += 1
			if index_word == self.args.top_wrong_words_restoration:
				break

		print('precision per character: ')
		
		for key, values in precision_chars.items():
			if values[1] != 0:
				p = values[0] / values[1]
				print(key + ': ' + str(p))
				precision_chars[key] = p
		

		print('recall per character: ')

		for key, values in recall_chars.items():
			if values[1] != 0:
				r = values[0] / values[1]
				print(key + ': ' + str(r))
				recall_chars[key] = r
		
		print('F1 measure per character: ')
		for key, r in recall_chars.items():
				p = precision_chars[key]
				if p != 0 and r != 0:
					f1 = 2 * p * r / (p + r)
					print(key + ': ' + str(f1))

		(char_acc, word_acc) = (correct_predicted_chars / (correct_predicted_chars + wrong_predicatd_chars), correct_predicted_words / total_words)
		print("char acc: " + str(char_acc) + ", word accuracy: " + str(word_acc) + ' ')
		return char_acc, word_acc

	def set_up_folders_saved_models(self):
		full_path_dir = self.args.folder_saved_model_per_epoch
		if self.args.save_model == True:
			if os.path.exists(full_path_dir) == False:
				os.makedirs(full_path_dir)
			elif self.args.load_model_name is None:
				print('a folder with the same name (' + self.args.folder_saved_model_per_epoch +\
					') for saving model already exists, delete it to continue or give other name to the folder of the saved model')
				exit(0)

	def parse_args(self):
		# when specify -load, specify also the nr of classes and if the model uses chars, words, sent
		# to restore diacritics run: 
		# python model_diacritice.py -buff 1000 -no_fast -load saved_models_diacritice/chars16-32-4classes -no_word -no_sent -classes 4 -no_dep -no_tag -restore raw_text.txt
		
		parser = argparse.ArgumentParser(description='Run diacritics model')
		parser.add_argument('-s', dest="save_model", action='store_false', default=True,\
							help="save the model (and weights), default=true")
		parser.add_argument('-f', dest="folder_saved_model_per_epoch",\
							action='store', default="char_word_sentence",\
							help="name of the folder to store the weights, default: char_word_sentence")
		parser.add_argument('-c', dest="corpus_rowiki",\
							action='store_true', default=False,\
							help="if you want to use rowiki corpus, beside parliament corpus, default=false")
		parser.add_argument('-test', dest="do_test",\
							action='store_true', default=False,\
							help="if you want to run test dataset, default=false")
		parser.add_argument('-n_test', dest="number_samples_test",\
							action='store', default=self.samples_number['par_test'] // self.batch_size, type=int,\
							help="number of samples for test accuracy, if -test is not set \
							this does not have any effect, default=100000")
		parser.add_argument('-e', dest="epochs",\
							action='store', default=20, type=int,\
							help="number of epochs, default=20")
		parser.add_argument('-r', dest="reset_iterators_every_epochs",\
							action='store', default=10, type=int,\
							help="reset the iterators for the dataset every nr epochs, default=10")
		parser.add_argument('-buff', dest="buffer_size_shuffle",\
							action='store', default=100000, type=int,\
							help="size of the buffer for shuffle, default=100000")
		parser.add_argument('-no_fast', dest="use_dummy_word_embeddings",\
							action='store_true', default=False,\
							help="use dummy word embeddings instead of fasttext, default=false")
		parser.add_argument('-load', dest="load_model_name",\
							action='store', default=None,\
							help="load presaved model and weights\
							, specify just the folder name, it will take the last epoch file,\
							default=None")
		parser.add_argument('-no_train_valid', dest="run_train_validation",\
							action='store_false', default=True,\
							help="run train and validation, if false you should set -load model param\
							, default=True")
		parser.add_argument('-mgpu', dest="percent_memory_gpu",\
							action='store', default=0.2, type=float,\
							help="percentage of the gpu memory to use, default=0.2")
		parser.add_argument('-wrong', dest="top_wrong_words_restoration",\
							action='store', default=30, type=int,\
							help="hardest words to restore, default=30")
		parser.add_argument('-no_char', dest="use_window_characters",\
							action='store_false', default=True,\
							help="if model should use window of characters, default=True")
		parser.add_argument('-no_word', dest="use_word_embedding",\
							action='store_false', default=True,\
							help="if model should use word embeddings, default=True")
		parser.add_argument('-no_sent', dest="use_sentence_embedding",\
							action='store_false', default=True,\
							help="if model should use sentence embedding, default=True")
		parser.add_argument('-no_tags', dest="use_tags",\
							action='store_false', default=True,\
							help="if model should use tags of the words, default=True")
		parser.add_argument('-no_deps', dest="use_deps",\
							action='store_false', default=True,\
							help="if model should use dependencies of the words, default=True")
		parser.add_argument('-hidden', dest="hidden_neurons",\
							action='append', type=int,\
							help="number of neurons on the hidden layer, no default")
		parser.add_argument('-classes', dest="nr_classes", default=4,\
							action='store', type=int,\
							help="number of classes to be used (4 or 5), default=4")
		parser.add_argument('-restore', dest="restore_diacritics", 
							action='store', default=None,\
							help="name of the file to restore diacritics")
		parser.add_argument('-batch_size_restore', dest="batch_size_restore", 
							action='store', type=int, default=64,\
							help="batch size ised for restoration")
		parser.add_argument('-output_file_restore', dest="output_file_restore", 
							action='store',  default='tmp_res.txt',\
							help="output of the file with diacritics")
		self.args = parser.parse_args()
		self.args.folder_saved_model_per_epoch += '/'

		
		if self.args.restore_diacritics is not None:
			# by default use the best model
			
			self.args.save_model = False
			self.args.do_test = False
			self.args.buffer_size_shuffle = 100
			self.args.run_train_validation = False

		if self.args.nr_classes != 4 and self.args.nr_classes != 5:
			print('classes has to be either 4 or 5, exit')
			exit(0)

		# for k in self.args.__dict__:
		# 	if self.args.__dict__[k] is not None:
		# 		print(k, '->', self.args.__dict__[k])

	def get_number_samples(self):
		if self.args.corpus_rowiki == False:
			inp_batches_train = self.samples_number['par_train'] // self.batch_size
			inp_batches_test = self.samples_number['par_test'] // self.batch_size
			inp_batches_valid = self.samples_number['par_valid'] // self.batch_size
		else:
			inp_batches_train = self.samples_number['full_train'] // self.batch_size
			inp_batches_test = self.samples_number['full_test'] // self.batch_size
			inp_batches_valid = self.samples_number['full_valid'] // self.batch_size
		return inp_batches_train, inp_batches_test, inp_batches_valid

	def get_input_list(self, characters_bi_lstm_layer, word_embeddings_layer, sentence_bi_lstm_layer, tags, deps):
		
		input_list = []

		if self.args.use_window_characters == True:
			input_list.append(characters_bi_lstm_layer)

		if self.args.use_word_embedding == True:
			input_list.append(word_embeddings_layer)

		if self.args.use_sentence_embedding == True:
			input_list.append(sentence_bi_lstm_layer)
		
		if self.args.use_tags == True:
			input_list.append(tags)

		if self.args.use_deps == True:
			input_list.append(deps)

		if len(input_list) == 1:
			return input_list[0]

		return input_list

	# class AttentionLayer(keras.layers.Layer):

	# 	def __init__(self, **kwargs):
	# 		super(AttentionLayer, self).__init__(**kwargs)

	# 	# multiple inputs represented as a list (of tuples?)
	# 	def build(self, input_shape):
	# 		assert isinstance(input_shape, list)
	# 		# Create a trainable weight variable for this layer.
	# 		# 
	# 		self.kernel = self.add_weight(name='kernel',
	# 									shape=(self.character_embedding_size, self.characters_cell_size * 2),
	# 									initializer='uniform',
	# 									trainable=True)
	# 		super(AttentionLayer, self).build(input_shape)  # Be sure to call this at the end

	# 	def call(self, x):
	# 		assert isinstance(x, list)
	# 		# W = windows size
	# 		# C = cell size
	# 		# E = embedding size
	# 		# inputt has shape (self.batch_size, W, E) 
	# 		# hidden has shape (self.batch_size, W, C * 2)
	# 		# kernel has shape (E, C * 2)
	# 		inputt, hidden = x
	# 		# inputt * kernel * hidden_states.T
	# 		return K.batch_dot(K.batch_dot(inputt, K.repeat_elements(K.expand_dims(self.kernel, axis=0), self.batch_size, axis=0)),\
	# 				K.permute_dimensions(hidden, (0, 2, 1)))
	# 		# res i j = correlation between input i and hidden state j

	# 	def compute_output_shape(self, input_shape):
	# 		assert isinstance(input_shape, list)
	# 		return (self.batch_size, self.window_character * 2 + 1, self.window_character * 2 + 1)

	# construct the model 
	def construct_model(self, sess):
		
		last_epoch = 0
		if self.args.load_model_name is not None:
			
			# folder_path_with_epochs = self.args.load_model_name 
			# epochs_files = os.listdir(folder_path_with_epochs)
			# sorted_epochs_files = sorted(epochs_files)
			# self.args.load_model_name = os.path.join(self.args.load_model_name, sorted_epochs_files[-1])
			# print('load model name: {}'.format(self.args.load_model_name))
			model = keras.models.load_model(self.args.load_model_name)
			#last_epoch = len(sorted_epochs_files)
		else:
			vocabulary_size = self.max_unicode_allowed + 1
			# character window 
			input_character_window = keras.layers.Input(shape=(self.window_character * 2 + 1,))
			character_embeddings_layer = keras.layers.Embedding(\
										input_dim=vocabulary_size,\
										output_dim=self.character_embedding_size)(input_character_window)

			character_lstm_layer = keras.layers.LSTM(
									units=self.characters_cell_size,\
									input_shape=(self.window_character * 2 + 1, self.character_embedding_size,),
									return_sequences=False)
			
			characters_bi_lstm_layer = keras.layers.Bidirectional(
									layer=character_lstm_layer,\
									merge_mode="concat")(character_embeddings_layer)
			# word token					
			word_embeddings_layer = keras.layers.Input(shape=(self.word_embedding_size,))

			# sentence token
			sentence_embeddings_layer = keras.layers.Input(shape=((self.window_sentence * 2 + 1, self.word_embedding_size,)))
			sentence_lstm_layer = keras.layers.LSTM(units=self.sentence_cell_size,\
													input_shape=(self.window_sentence * 2 + 1, self.word_embedding_size,))	

			sentence_bi_lstm_layer = keras.layers.Bidirectional(layer=sentence_lstm_layer,\
																merge_mode="concat")(sentence_embeddings_layer)
			# tags 
			tags = keras.layers.Input(shape=(1,),\
									dtype='float32')
			# deps
			deps = keras.layers.Input(shape=(1,),\
									dtype='float32')
			# merged
			input_list = self.get_input_list(characters_bi_lstm_layer,\
				word_embeddings_layer, sentence_bi_lstm_layer, tags, deps)
			if len(input_list) > 1:
				merged_layer = keras.layers.concatenate(input_list, axis=-1)
			else:
				merged_layer = input_list[0]
			prev_layer = merged_layer
			#prev_layer = keras.layers.Flatten()(prev_layer)
			#attention = AttentionLayer()([character_embeddings_layer, prev_layer])
			#attention = keras.layers.Flatten()(attention)
			attention = prev_layer

			# hidden layers
			for h_neurons in self.args.hidden_neurons:
				attention = keras.layers.Dense(h_neurons, activation='tanh')(attention)
			output = keras.layers.Dense(self.args.nr_classes, activation='softmax')(attention)

			# [input_character_window, word_embeddings_layer, sentence_embeddings_layer]
			model_input = self.get_input_list(input_character_window, word_embeddings_layer, sentence_embeddings_layer,\
				tags, deps)

			model = keras.models.Model(inputs=model_input,\
									outputs=output)
			model.compile(optimizer='adam',\
						loss='categorical_crossentropy',\
						metrics=['accuracy'])
			print(model.summary())
		return model, last_epoch

	def run_task(self):
		inp_batches_train, inp_batches_test, inp_batches_valid = self.get_number_samples()
		self.set_up_folders_saved_models()

		if self.args.use_dummy_word_embeddings == False:
			self.model_embeddings = FastTextWrapper.load_fasttext_format(self.fast_text)

		config = tf.ConfigProto()
		config.gpu_options.per_process_gpu_memory_fraction = self.args.percent_memory_gpu

		with tf.Session(config=config) as sess:

			if inp_batches_test < self.args.number_samples_test:
				print("cannot start, too many test samples given, has to be lower than "\
						+ str(inp_batches_test))

			print("char, word, sentence - char cell:{}, word cell: {}, hidden: {}"\
				.format(self.characters_cell_size, self.sentence_cell_size, self.neurons_dense_layer_after_merge))

			model, last_epoch = self.construct_model(sess)
			# run test and validation 
			if self.args.run_train_validation == True:

				dt_train = self.get_dataset(self.train_files, sess)
				dt_valid = self.get_dataset(self.valid_files, sess)
				iterator_train = dt_train.make_initializable_iterator()
				iterator_valid = dt_valid.make_initializable_iterator()

				for i in range(last_epoch, last_epoch + self.args.epochs):

					print('epoch: ' + str(i + 1))
					# reset iterators
					if (i - last_epoch) % self.args.reset_iterators_every_epochs == 0:
						print('resseting iterators')
						sess.run(iterator_valid.initializer)
						valid_inp, valid_out = iterator_valid.get_next()
						valid_char_window, valid_words, valid_sentence, valid_tags, valid_deps = valid_inp

						sess.run(iterator_train.initializer)
						train_inp, train_out = iterator_train.get_next()
						train_char_window, train_words, train_sentence, train_tags, train_deps = train_inp
					
					# train an epoch
					train_input = self.get_input_list(train_char_window, train_words, train_sentence,\
						train_tags, train_deps)
					model.fit(\
						train_input,\
						[train_out],\
						steps_per_epoch=inp_batches_train//self.args.reset_iterators_every_epochs,\
						epochs=1,\
						verbose=1)

					# save weights
					if self.args.save_model == True:
						print('saving model (and weights)')
						full_path_epoch_weights = os.path.join(self.args.folder_saved_model_per_epoch, 'epoch_' + str(i) + '.h5')
						model.save(full_path_epoch_weights)
					# validate 
					valid_input = self.get_input_list(valid_char_window, valid_words, valid_sentence,
						valid_tags, valid_deps)
					[valid_loss, valid_acc] = model.evaluate(valid_input,\
												valid_out,\
												verbose=1,\
												steps=inp_batches_valid//self.args.reset_iterators_every_epochs)
					print("validation - loss: " + str(valid_loss) +  " acc: " + str(valid_acc))
			# test
			if self.args.do_test:
				self.compute_test_accuracy(sess, model)
			if self.args.restore_diacritics:
				self.restore_diacritics(sess, model)

if __name__ == "__main__":
	diacritics = Diacritics()
	diacritics.args.use_window_characters = True
	diacritics.args.use_dummy_word_embeddings = True
	diacritics.args.use_sentence_embedding = False
	diacritics.args.use_word_embedding = False
	diacritics.args.buffer_size_shuffle = 1000 
	diacritics.args.load_model_name = './saved_models_diacritice/chars16-32-4classes/'
	diacritics.args.nr_classes = 4
	diacritics.args.use_tags = False
	diacritics.args.use_deps = False
	diacritics.args.restore_diacritics = 'book.txt'
	diacritics.args.batch_size_restore = 1024

	diacritics.args.save_model = False
	diacritics.args.do_test = False
	diacritics.args.run_train_validation = False
	diacritics.args.output_file_restore = 'tmp_res2.txt'

	diacritics.run_task()



from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.complexity.word.name_entity_enum import NamedEntityONEnum 
from rb.core.text_element_type import TextElementType
from typing import List, Callable
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class NoNamedEntity(ComplexityIndex):

    """named_ent_type depends on the model, there are 2 enum options
        see name_entity_enum.py for those"""
    def __init__(self, lang: Lang, named_ent_type: NamedEntityONEnum,
            reduce_depth: int, reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.WORD,
                                 abbr="NoNamedEntity", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)
        self.named_ent_type = named_ent_type

    def process(self, element: TextElement) -> float:
        return self.reduce_function(self.compute_above(element))

    def compute_below(self, element: TextElement) -> float:
        if element.is_word() == True:
            if element.ent_type_ != 0 and element.ent_type_ == self.named_ent_type.name:
                return 1
            else:
                return 0
        elif element.depth <= self.reduce_depth:
            res = 0
            for child in element.components:
                res += self.compute_below(child)
            return res
    
    def compute_above(self, element: TextElement) -> List[float]:
        if element.depth > self.reduce_depth:
            values = []
            for child in element.components:
                values += self.compute_above(child)
            element.indices[self] = self.reduce_function(values)
        elif element.depth == self.reduce_depth:
            values = [self.compute_below(element)]
            element.indices[self] = self.reduce_function(values)
        else:
            logger.error('wrong reduce depth value.')
        return values
    
    def __repr__(self):
        return self.reduce_function_abbr + self.reduce_depth_abbr + self.abbr + "_" + self.named_ent_type.name.lower()


from rb.comprehension.utils.graph.cm_graph_do import CmGraphDO
from rb.comprehension.utils.graph.cm_edge_do import CmEdgeDO
from rb.comprehension.utils.graph.cm_node_do import CmNodeDO

import networkx as nx


class GraphMetrics():

    def __init__(self, graph: CmGraphDO):
        self.cm_graph = graph
        self.build_networkx_graph()
    

    def build_networkx_graph(self):
        self.G = nx.Graph()

        for node in self.cm_graph.node_list:
            if node.is_active():
                self.G.add_node(node.word.lemma)

        for edge in self.cm_graph.edge_list:
            if edge.is_active():
                self.G.add_edge(edge.node1, edge.node2, weight=edge.score)
    

    def degree_centrality(self):
        return nx.algorithms.centrality.degree_centrality(self.G)


    def clossness_centrality(self):
        return nx.algorithms.centrality.closeness_centrality(self.G)


    def betweenness_centrality(self):
        return nx.algorithms.centrality.betweenness_centrality(self.G)


    def harmonic_centrality(self):
        return nx.algorithms.centrality.harmonic_centrality(self.G)


from rb.core.lang import Lang
from rb.core.document import Document
from rb.complexity.complexity_index import ComplexityIndex, compute_indices
from rb.complexity.cohesion.adj_cohesion import AdjCohesion
from rb.similarity.word2vec import Word2Vec
from rb.similarity.vector_model import VectorModelType, CorporaEnum, VectorModel
from rb.similarity.vector_model_factory import VECTOR_MODELS, create_vector_model
from rb.cna.cna_graph import CnaGraph
from typing import Tuple, List
from sklearn.svm import SVR
import pickle
import os
import csv
from rb.utils.rblogger import Logger
from typing import List, Tuple
import numpy as np


logger = Logger.get_logger()


class Fluctuations:

    def __init__(self):
        pass
    
    def get_vector_model(self, lang: Lang = Lang.RO) -> VectorModel:
        global logger
        if lang is Lang.RO:
            vector_model = create_vector_model(Lang.RO, VectorModelType.from_str('word2vec'), "readme")
        elif lang is Lang.EN:
            vector_model = create_vector_model(Lang.EN, VectorModelType.from_str("word2vec"), "coca")
        else:
            logger.error(f'Language {lang.value} is not supported for fluctuations task')
            vector_model = None
        return vector_model

    def compute_thresholds(self, values: List[float]) -> Tuple[int, int]:
        if len(values) > 1:
            stdev = np.std(values)
            avg = np.mean(values)
        elif len(values) == 1:
            avg = values[0]
            stdev = 1
        else:
            avg = -1
            stdev = -1
        return (max(0, avg + 2.0 * stdev), max(0, avg - 2.0 * stdev))


    def compute_indices(self, text: str, lang: Lang) -> List[List]:
        doc = Document(lang=lang, text=text)
        vector_model = self.get_vector_model(lang=lang)
        cna_graph = CnaGraph(doc=doc, models=[vector_model])
        compute_indices(doc=doc, cna_graph=cna_graph)

        indices_sent = {
                        'AvgSentUnqPOSMain_noun': 
                            {Lang.RO: 'Numărul de substantive unice per propoziție',
                             Lang.EN: 'Number of unique nouns per sentence'}, 
                        'AvgSentUnqPOSMain_verb': 
                            {Lang.RO: 'Numărul de verbe unice per propoziție',
                             Lang.EN: 'Number of unique verbs per sentence'},
                        'AvgSentUnqPOSMain_adj':
                            {Lang.RO: 'Numărul de adjective unice per propoziție',
                             Lang.EN: 'Number of unique adjectives per sentence'},
                        'AvgSentUnqPOSMain_adv':
                            {Lang.RO: 'Numărului de adverbe unice per propoziție',
                             Lang.EN: 'Number of unique adverbs per sentence'},
                        'AdjExtCoh_SENT':
                            {Lang.RO: 'Coeziunea propoziției curente cu propozițiile vecine',
                             Lang.EN: 'Cohesion of the current sentence with its neighbouring sentences'}
                        }
        indices_block = {
                        'AvgSentUnqPOSMain_noun': 
                            {Lang.RO: 'Media numărului de substantive unice per frază',
                             Lang.EN: 'Average of the number of unique nouns per paragraph'}, 
                        'AvgSentUnqPOSMain_verb': 
                            {Lang.RO: 'Media numărului de verbe unice per frază',
                             Lang.EN: 'Average of the number of unique verbs per paragraph'},
                        'AvgSentUnqPOSMain_adj':
                            {Lang.RO: 'Media numărului de adjective unice per frază',
                             Lang.EN: 'Average of the number of unique adjectives per paragraph'},
                        'AvgSentUnqPOSMain_adv':
                            {Lang.RO: 'Media numărului de adverbe unice per frază',
                             Lang.EN: 'Average of the number of unique adverbs per paragraph'},
                        'AdjExtCoh_BLOCK':
                            {Lang.RO: 'Coeziunea paragrafului curent cu paragrafele vecine',
                             Lang.EN: 'Cohesion of the current paragraph with its neighbouring paragraphs'}
                        }
        result = []

        for ind_sent, _ in indices_sent.items():
            d = {'index': ind_sent, 'index_description': indices_sent[ind_sent][lang],
                 'level': 'sentence', 'values': [], 'text': []}
            for sent in doc.get_sentences():
                for key, v in sent.indices.items():
                    if repr(key) == ind_sent:
                        d['values'].append(v)
                        d['text'].append(sent.text)
            maxt, mint = self.compute_thresholds(d['values'])
            d['threshold'] = {
                'min': str(mint),
                'max': str(maxt)
            }
            for i, v in enumerate(d['values']):
                d['values'][i] = str(v)
            result.append(d)
        
        for ind_block, _ in indices_block.items():
            d = {'index': ind_block, 'index_description': indices_block[ind_block][lang],
                 'level': 'paragraph', 'values': [], 'text': []}
            for block in doc.get_blocks():
                for key, v in block.indices.items():
                    if repr(key) == ind_block:
                        d['values'].append(v)
                        d['text'].append(block.text)
            maxt, mint = self.compute_thresholds(d['values'])
            d['threshold'] = {
                'min': str(mint),
                'max': str(maxt)
            }

            for i, v in enumerate(d['values']):
                d['values'][i] = str(v)
            result.append(d)
        return result


from rb.core.lang import Lang
from rb.core.document import Document
from rb.complexity.complexity_index import ComplexityIndex, compute_indices
from rb.similarity.word2vec import Word2Vec
from rb.similarity.vector_model import VectorModelType, CorporaEnum, VectorModel
from rb.similarity.vector_model_factory import VECTOR_MODELS, create_vector_model
from typing import Tuple, List
from sklearn.svm import SVR
import pickle
import os
import csv
from werkzeug import secure_filename
import uuid
from rb.cna.cna_graph import CnaGraph
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class EssayScoring:


    def __init__(self):
        pass

    def get_vector_model(self, lang: Lang = Lang.RO) -> VectorModel:
        global logger
        if lang is Lang.RO:
            vector_model = create_vector_model(Lang.RO, VectorModelType.from_str('word2vec'), "readme")
        elif lang is Lang.EN:
            vector_model = create_vector_model(Lang.EN, VectorModelType.from_str("word2vec"), "coca")
        else:
            logger.error(f'Language {lang.value} is not supported for essay scoring task')
            vector_model = None
        return vector_model
    
    """ given a csv, transform it into text files and a csv with filename, grade"""
    def create_files_from_csv(self, path_to_csv_file: str, path_to_folder=".", output_stats_csv='stats.csv'):
        essay_r = csv.reader(open(path_to_csv_file, 'rt', encoding='utf-8'))
        
        csv_stats = [['filename', 'grade']]
        try:
            os.mkdir(path_to_folder)
            logger.info("Directory {} created".format(path_to_folder))
        except FileExistsError:
            logger.info("Directory {} already exists".format(path_to_folder))
        
        for i, row in enumerate(essay_r):
            try:
                grade = float(row[0])
                content = str(row[1])
            except:
                continue
            text_file = secure_filename(str(uuid.uuid4()) + '.txt')
            full_path_text_file = os.path.join(path_to_folder, text_file)
            entry_csv = [text_file, grade]
            csv_stats.append(entry_csv)
            with open(full_path_text_file, 'wt', encoding='utf-8') as f:
                f.write(content)

        with open(os.path.join(path_to_folder, output_stats_csv), 'wt', encoding='utf-8') as csv_essays_eval:
            csv_writer = csv.writer(csv_essays_eval)
            csv_writer.writerows(csv_stats)


    def compute_indices(self, base_folder: str='essays_ro', write_file: str='measurements.csv',
            stats: str='stats.csv', lang: Lang = Lang.RO, nr_docs: int=None) -> List[List]:
        
        indices = []
        if lang is Lang.RO:
            with open('rb/processings/scoring/indices_ro_scoring.txt', 'rt', encoding='utf-8') as f:
                for line in f:
                    indices.append(line.strip())
                    
        vector_model = self.get_vector_model(lang=lang)
            
        all_rows = []
        first_row = ['filename', 'grade']
        essay_r = csv.reader(open(os.path.join(base_folder, stats), 'rt', encoding='utf-8'))

        for i, row in enumerate(essay_r):
            if i == 0:  continue
            if (not (nr_docs is None)) and i > nr_docs:  break
            logger.info('Computing indices for document number {}'.format(i))
            try:
                grade = float(row[1])
                file_name = str(row[0])
            except:
                continue

            with open(os.path.join(base_folder, file_name), 'rt', encoding='utf-8') as f:
                content = f.read()
            
            doc = Document(lang=lang, text=content)
            cna_graph = CnaGraph(doc=doc, models=[vector_model])
            compute_indices(doc=doc, cna_graph=cna_graph)

            row = []
            if len(first_row) == 2:
                for ind in indices:
                    for key, v in doc.indices.items():
                        if str(key) == ind:
                            first_row.append(key)
                all_rows.append(first_row)

            row = [file_name, grade]
            for ind in indices:
                for key, v in doc.indices.items():
                    if str(key) == ind:
                        row.append(v)
            
            if len(row) == len(first_row):
                all_rows.append(row)

        with open(os.path.join(base_folder, write_file), 'wt', encoding='utf-8') as csv_essays_eval:
            csv_writer = csv.writer(csv_essays_eval)
            csv_writer.writerows(all_rows)
        return all_rows
    
    def read_indices(self, base_folder: str = 'essays_ro',
         path_to_csv_file='measurements.csv') -> List[List[float]]:
        essay_r = csv.reader(open(os.path.join(base_folder, path_to_csv_file), 'rt', encoding='utf-8'))
        results = []
        """ first row is the score """
        for i, row in enumerate(essay_r):
            if i == 0:  continue
            results.append(row[1:])
        return results

    def train_svr(self, results: List[List], save_model_file=None):
        total = len(results)
        train_samples = int(total * 0.80)

        train = results[:train_samples]
        test = results[train_samples:]

        y = [float(r[0]) for r in train]
        X = [r[1:] for r in train]

        clf = SVR(gamma='scale', C=1.0, epsilon=0.2)
        clf.fit(X, y)
        if save_model_file:
            pickle.dump(clf, open(save_model_file, 'wb'))

        loss = 0
        grades, dev_in = [], []

        for sample_x in test:
            grades.append(float(sample_x[0]))
            X = sample_x[1:]
            dev_in.append(X)
        res = clf.predict(dev_in)

        for r, grade in zip(res, grades):            
            loss += abs(r - grade)
        loss /= len(res)

        logger.info('Scoring loss: {}'.format(loss))

    def predict(self, content: str, file_to_svr_model: str, lang=Lang.RO) -> float:

        svr = pickle.load(open(file_to_svr_model, "rb"))

        doc = Document(lang=lang, text=content)
        vector_model = self.get_vector_model(lang=lang)
        cna_graph = CnaGraph(doc=doc, models=[vector_model])
        compute_indices(doc=doc, cna_graph=cna_graph)

        indices = []
        if lang is Lang.RO:
            with open('rb/processings/scoring/indices_ro_scoring.txt', 'rt', encoding='utf-8') as f:
                for line in f:
                    indices.append(line.strip())
        values_indices = []
        for ind in indices: 
            for key, v in doc.indices.items():
                if repr(key) == ind:
                    values_indices.append(v)

        grade = svr.predict([values_indices])[0]
        return max(grade - 7, 0.1)

        


from typing import List

from rb.core.lang import Lang
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType
from rb.complexity.measure_function import MeasureFunction
from rb.core.pos import POS as PosEum
from rb.cna.cna_graph import CnaGraph


# create all indices
# dependencies need to be putted in function because otherwise circular dependencies happens
def create(lang: Lang, cna_graph: CnaGraph) -> List["ComplexityIndex"]:
    from rb.complexity.syntax.dep_enum import DepEnum
    from rb.complexity.morphology.pos_main import PosMain
    from rb.complexity.morphology.unq_pos_main import UnqPosMain
    from rb.complexity.morphology.pron_type_enum import PronounTypeEnum
    from rb.complexity.morphology.pron import Pronoun
    
    indices = []

    indices.append(PosMain(lang, PosEum.NOUN, TextElementType.SENT.value, MeasureFunction.AVG))
    indices.append(PosMain(lang, PosEum.VERB, TextElementType.SENT.value, MeasureFunction.AVG))
    indices.append(PosMain(lang, PosEum.ADJ, TextElementType.SENT.value, MeasureFunction.AVG))
    indices.append(PosMain(lang, PosEum.ADV, TextElementType.SENT.value, MeasureFunction.AVG))
    indices.append(PosMain(lang, PosEum.PRON, TextElementType.SENT.value, MeasureFunction.AVG))

    indices.append(UnqPosMain(lang, PosEum.NOUN, TextElementType.SENT.value, MeasureFunction.AVG))
    indices.append(UnqPosMain(lang, PosEum.VERB, TextElementType.SENT.value, MeasureFunction.AVG))
    indices.append(UnqPosMain(lang, PosEum.ADJ, TextElementType.SENT.value, MeasureFunction.AVG))
    indices.append(UnqPosMain(lang, PosEum.ADV, TextElementType.SENT.value, MeasureFunction.AVG))
    indices.append(UnqPosMain(lang, PosEum.PRON, TextElementType.SENT.value, MeasureFunction.AVG))


    indices.append(PosMain(lang, PosEum.NOUN, TextElementType.BLOCK.value, MeasureFunction.AVG))
    indices.append(PosMain(lang, PosEum.VERB, TextElementType.BLOCK.value, MeasureFunction.AVG))
    indices.append(PosMain(lang, PosEum.ADJ, TextElementType.BLOCK.value, MeasureFunction.AVG))
    indices.append(PosMain(lang, PosEum.ADV, TextElementType.BLOCK.value, MeasureFunction.AVG))
    indices.append(PosMain(lang, PosEum.PRON, TextElementType.BLOCK.value, MeasureFunction.AVG))

    indices.append(UnqPosMain(lang, PosEum.NOUN, TextElementType.BLOCK.value, MeasureFunction.AVG))
    indices.append(UnqPosMain(lang, PosEum.VERB, TextElementType.BLOCK.value, MeasureFunction.AVG))
    indices.append(UnqPosMain(lang, PosEum.ADJ, TextElementType.BLOCK.value, MeasureFunction.AVG))
    indices.append(UnqPosMain(lang, PosEum.ADV, TextElementType.BLOCK.value, MeasureFunction.AVG))
    indices.append(UnqPosMain(lang, PosEum.PRON, TextElementType.BLOCK.value, MeasureFunction.AVG))
    
    if lang is Lang.RO or lang is Lang.EN:
        for pt in PronounTypeEnum:
            indices.append(Pronoun(lang, pt, TextElementType.SENT.value, MeasureFunction.AVG))
            indices.append(Pronoun(lang, pt, TextElementType.BLOCK.value, MeasureFunction.AVG))

    return indices


from typing import Dict

from rb.comprehension.utils.graph.cm_graph_do import CmGraphDO
from rb.comprehension.utils.graph.cm_node_do import CmNodeDO

ActivationMap = Dict[CmNodeDO, float]


class PageRank():

    def __init__(self):
        self.max_iter = 100000
        self.eps = 0.0001
        self.prob = 0.85

    
    def run_page_rank(self, graph: CmGraphDO) -> None:
        page_rank_values = graph.get_activation_map()
        new_page_rank_values = self.run_page_rank_with_activation(page_rank_values, graph)

        for key, value in new_page_rank_values.items():
            node = graph.get_node(key)
            node.set_activation_score(value)


    def run_page_rank_with_activation(self, activation_map: ActivationMap, graph: CmGraphDO) -> ActivationMap:
        iteration = 0
        current_page_rank_values = dict(activation_map)
        while iteration < self.max_iter:
            r = self.calculateR(current_page_rank_values, graph)
            temp_page_rank_values = {}
            done = True

            for node in graph.node_list:
                temp_pr_value = self.compute_temp_page_rank_values(current_page_rank_values, graph, node, r)
                prev_pr_value = self.get_page_rank_value(current_page_rank_values, node, graph)

                temp_page_rank_values[node] = temp_pr_value

                if (temp_pr_value - prev_pr_value) / prev_pr_value >= self.eps:
                    done = False
            
            current_page_rank_values = temp_page_rank_values
            if done:
                break
            
            iteration += 1
        
        return current_page_rank_values


    def calculateR(self, page_rank_values: ActivationMap, graph: CmGraphDO) -> float:
        r = 0
        n = len(graph.node_list)
        for node in graph.node_list:
            node_edge_list = graph.get_activate_edges_for_node(node)
            node_degree = len(node_edge_list)
            node_page_rank_value = self.get_page_rank_value(page_rank_values, node, graph)
            if node_degree > 0:
                r += (1.0 - self.prob) * (node_page_rank_value / n)
            else:
                r += (node_page_rank_value / n)
        return r


    def compute_temp_page_rank_values(self, page_rank_values: ActivationMap,
                                        graph: CmGraphDO, node: CmNodeDO, r: float) -> float:
        res = r
        node_edge_list = graph.get_activate_edges_for_node(node)
        for edge in node_edge_list:
            neighbour = edge.get_opposite_node(node)
            neighbour_edge_list = graph.get_activate_edges_for_node(neighbour)
            normalize =  len(neighbour_edge_list)
            res += self.prob * (self.get_page_rank_value(page_rank_values, neighbour, graph) / normalize)

        return res


    def get_page_rank_value(self, page_rank_values: ActivationMap, node: CmNodeDO, graph: CmGraphDO) -> float:
        if node in page_rank_values.keys():
            return page_rank_values[node]
        
        return 1 / len(graph.node_list)



from rb.core.lang import Lang
from rb.core.sentence import Sentence
from rb.core.word import Word
from rb.core.document import Document

from rb.comprehension.utils.distance_strategies.cm_syntactic_graph import CmSyntacticGraph

from copy import deepcopy

from typing import Tuple, List
Dependency = Tuple[Word, Word, str]
Dependencies = List[Dependency]


class CmCorefIndexer():

    def __init__(self, document: Document, lang: Lang):
        self.lang = lang
        self.document = document

    
    def get_cm_syntactic_graph(self, sentence: Sentence, sentence_index: int) -> CmSyntacticGraph:
        dependencies = sentence.get_dependencies()
        syntactic_graph = CmSyntacticGraph()

        for index, dependency in enumerate(dependencies):
            dependent_word = self.get_actual_word(dependency[0], sentence_index)
            governor_word = self.get_actual_word(dependency[1], sentence_index)

            if dependent_word.is_content_word() and governor_word.is_content_word():
                syntactic_graph.index_edge(dependent_word, governor_word)
                continue

            if not dependent_word.is_content_word() and governor_word.is_content_word():
                content_word = governor_word
                non_content_word = dependent_word
            elif dependent_word.is_content_word() and not governor_word.is_content_word():
                content_word = dependent_word
                non_content_word = governor_word
            else:
                continue

            clone = sentence.get_dependencies()
            del clone[index]
            self.explore(content_word, non_content_word, clone, syntactic_graph, sentence_index)

        return syntactic_graph
            

    def explore(self, content_word: Word, non_content_word: Word, dependencies: Dependencies,
                    syntactic_graph: CmSyntacticGraph, sentence_index: int) -> None:
        for index, dependency in enumerate(dependencies):
            dependent_word = self.get_actual_word(dependency[0], sentence_index)
            governor_word = self.get_actual_word(dependency[1], sentence_index)

            if dependent_word.lemma == non_content_word.lemma:
                if governor_word.is_content_word():
                    syntactic_graph.index_edge(content_word, governor_word)
                else:
                    clone = deepcopy(dependencies)
                    clone.remove(index)
                    self.explore(content_word, governor_word, clone, syntactic_graph, sentence_index)
            elif governor_word.lemma == non_content_word.lemma:
                if dependent_word.is_content_word():
                    syntactic_graph.index_edge(content_word, dependent_word)
                else:
                    clone = deepcopy(dependencies)
                    clone.remove(index)
                    self.explore(content_word, dependent_word, clone, syntactic_graph, sentence_index)
                

    # TODO use coref
    def get_actual_word(self, word: Word, sentence_index: int) -> Word:
        return word

    # TODO 
    def index_coreference(self):
        pass


from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoNumFormEnum(Enum):
    DIGIT = auto()
    ROMAN = auto()
    WORD = auto()


class RoPOSFeatureNumForm(POSFeature):


    _INSTANCE = None
    
    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.NRF,
                            values=None, pos_supported=None)
        self.values = set([ro_nr_form for _, ro_nr_form in RoNumFormEnum.__members__.items()])
        self.pos_supported = set([POS.NUM])

    @classmethod
    def get_instance(cls) -> "RoPOSFeatureNumForm":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeatureNumForm()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoNumFormEnum]:
        pattern = '(NumForm=[a-zA-Z,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        nr_forms = matches[0].split('=')[1].split(',')
        values = []
        for nr_form in nr_forms:
            if nr_form == 'Digit':
                values.append(RoNumFormEnum.DIGIT)
            elif nr_form == 'Roman':
                values.append(RoNumFormEnum.ROMAN)
            elif nr_form == 'Word':
                values.append(RoNumFormEnum.WORD)
        return values


import os

from rb.core.lang import Lang
from rb.similarity.vector_model import VectorModel, VectorModelType
from rb.utils.downloader import download_model, check_version


class LSA(VectorModel):

    def __init__(self, name: str, lang: Lang, dim: int = 300):
        VectorModel.__init__(self, VectorModelType.LSA, name, lang, dim)  
        


#check if predicative name is in the list with jobs, because it is right Fata este doctor.
def is_job_noun(spn, predicative_name_text, exceptions):
    if spn == ("3", "s", "f") or spn == ("-", "s", "f"):
        for job in exceptions["Jobs"]:
            if job.split(" ")[0].lower() == predicative_name_text.text.lower():
                return True
    return False


#if it is a collective subject, it is also correct Majoritatea au votat DA.
def is_collective_subject(subject, spv, exceptions):
	if subject.tag_[0] == "P":
		return False
	if subject.text.lower() in exceptions["Collective nouns"] and spv == ("3", "p"):
		return True
	return False

def are_describing_persons(subjects, exceptions):
	persons = []
	for subject in subjects:
		if subject.text.lower() in exceptions["Persons"]:
			persons.append(subject)
	return persons



""" Cohesion between current sentence/block and its neighbours. This is different from adj_cojesion
which averages/stdev cohesions on block or doc level. There is no avgstddev here.  """

from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType   
from typing import List, Callable
from rb.similarity.vector_model import VectorModel
from rb.cna.cna_graph import CnaGraph
from rb.cna.edge_type import EdgeType

from rb.utils.rblogger import Logger

logger = Logger.get_logger()

class AdjExternalCohesion(ComplexityIndex):

    def __init__(self, lang: Lang, element_type: TextElementType,
            reduce_depth: int = None, reduce_function: MeasureFunction = None,
            cna_graph: CnaGraph = None):
        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.COHESION,
                                 reduce_depth=reduce_depth, reduce_function=reduce_function,
                                 abbr="AdjExtCoh")
        self.element_type = element_type
        self.cna_graph = cna_graph   
        if not (element_type is TextElementType.BLOCK or element_type is TextElementType.SENT):
            logger.error('For index {} element_type has to {} or {}'.format(TextElementType.BLOCK, TextElementType.SENT))

    def process(self, element: TextElement) -> float:
        
        if self.element_type is TextElementType.SENT:
            elems = element.get_sentences()
        else:
            elems = element.get_blocks()

        if len(elems) == 0:
            return ComplexityIndex.IDENTITY
        elif len(elems) == 1:
            elems[0].indices[self] = ComplexityIndex.IDENTITY
            return ComplexityIndex.IDENTITY
        elif len(elems) == 2:
            sim_edge = self.cna_graph.edges(node=(elems[0], elems[1]), edge_type=EdgeType.SEMANTIC, 
                                            vector_model=None)
            v = sim_edge[0][2]
            elems[0].indices[self] = elems[1].indices[self] = v
            return ComplexityIndex.IDENTITY
        else:
            for i, elem in enumerate(elems):
                if i == 0:
                    sim_edge = self.cna_graph.edges(node=(elems[i], elems[i + 1]), edge_type=EdgeType.SEMANTIC,
                                                    vector_model=None)
                    v = sim_edge[0][2]
                    elems[i].indices[self] = v
                elif i == len(elems) - 1:
                    sim_edge = self.cna_graph.edges(node=(elems[i], elems[i - 1]), edge_type=EdgeType.SEMANTIC,
                                                    vector_model=None)
                    v = sim_edge[0][2]
                    elems[i].indices[self] = v
                else:
                    sim_edge = self.cna_graph.edges(node=(elems[i], elems[i - 1]), edge_type=EdgeType.SEMANTIC,
                                                    vector_model=None)
                    v = sim_edge[0][2]
                    sim_edge = self.cna_graph.edges(node=(elems[i], elems[i + 1]), edge_type=EdgeType.SEMANTIC,
                                                    vector_model=None)
                    v += sim_edge[0][2]
                    v /= 2
                    elems[i].indices[self] = v
            return ComplexityIndex.IDENTITY     
    
    def __repr__(self):
        return self.abbr + '_' + self.element_type.name


from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType
from rb.complexity.morphology.pron_type_enum import PronounTypeEnum
from typing import List, Callable, Set
from rb.core.pos import POS as PosEum
from rb.utils.rblogger import Logger
from typing import Dict, List
logger = Logger.get_logger()


class Pronoun(ComplexityIndex):
    
    
    pronoun_dict: Dict[PronounTypeEnum, List[str]] = None
    
    def __init__(self, lang: Lang, pron_type: PronounTypeEnum,
            reduce_depth: int, reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.MORPHOLOGY,
                                 abbr="Pronoun", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)
        if Pronoun.pronoun_dict is None:
            Pronoun.parse_pronoun_list(lang)

        self.pron_type = pron_type


    @staticmethod
    def parse_pronoun_list(lang: Lang):

        if lang is Lang.RO:
            path_to_wordlist = 'resources/ro/wordlists/pronouns_ro.txt'
        elif lang is Lang.EN:
            path_to_wordlist = 'resources/en/wordlists/pronouns_en.txt'

        with open(path_to_wordlist, 'rt', encoding='utf-8') as f:
            Pronoun.pronoun_dict = {}
            for pt in PronounTypeEnum:
                Pronoun.pronoun_dict[pt] = []

            pron_type = None
            for i, line in enumerate(f):
                if line.find('first_person') != -1:
                    pron_type = PronounTypeEnum.FST
                elif line.find('second_person') != -1:
                    pron_type = PronounTypeEnum.SND
                elif line.find('third_person') != -1:
                    pron_type = PronounTypeEnum.THRD
                elif line.find('interrogative') != -1:
                    pron_type = PronounTypeEnum.INT
                elif line.find('indefinite') != -1:
                    pron_type = PronounTypeEnum.INDEF
                elif len(line.strip()) == 0:
                    continue
                elif pron_type is not None:
                    Pronoun.pronoun_dict[pron_type].append(line.strip())

    def process(self, element: TextElement) -> float:
        return self.reduce_function(self.compute_above(element))

    def compute_below(self, element: TextElement) -> List[str]:
        if element.is_word() == True:
            res = []
            if element.text.lower() in Pronoun.pronoun_dict[self.pron_type]:
                res.append(element.text)
            return res
        elif element.depth <= self.reduce_depth:
            res = []
            for child in element.components:
                res += self.compute_below(child)
            return res

    def compute_above(self, element: TextElement) -> List[float]:
        if element.depth > self.reduce_depth:
            values = []
            for child in element.components:
                values += self.compute_above(child)
            element.indices[self] = self.reduce_function(values)
        elif element.depth == self.reduce_depth:
            element.indices[self] = len(self.compute_below(element))
            values = [len(self.compute_below(element))]
            element.indices[self] = self.reduce_function(values)
        else:
            logger.error('wrong reduce depth value.')
        return values

    def __repr__(self):
        return self.reduce_function_abbr + self.reduce_depth_abbr + self.abbr + "_" + self.pron_type.name.lower()



from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType   
from typing import List, Callable
from rb.similarity.vector_model import VectorModel
from rb.cna.cna_graph import CnaGraph
from rb.cna.edge_type import EdgeType

from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class IntraCohesion(ComplexityIndex):

    """IntraCohesion between elements element_type"""
    def __init__(self, lang: Lang, element_type: TextElementType,
            reduce_depth: int, reduce_function: MeasureFunction, cna_graph: CnaGraph):
        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.COHESION,
                                 reduce_depth=reduce_depth, reduce_function=reduce_function,
                                 abbr="IntraCoh")
        self.element_type = element_type
        self.cna_graph = cna_graph    
        if element_type.value > reduce_depth:
            logger.error('For index {} element_type has to be lower or equal than reduce_depth'.format(self))

    def process(self, element: TextElement) -> float:
        values = self.compute_above(element)
        return self.reduce_function(values) if len(values) > 0 else ComplexityIndex.IDENTITY

    def compute_below(self, element: TextElement) -> List[float]:
        if element.depth == self.element_type.value:
            sim_values = []
            part_of_edges = self.cna_graph.edges(node=element, edge_type=EdgeType.PART_OF, vector_model=None)
            for edge in part_of_edges:
                sim_edge = self.cna_graph.edges(node=(edge[0], edge[1]), edge_type=EdgeType.SEMANTIC, vector_model=None)
                if len(sim_edge) > 0 and len(sim_edge[0]) >= 3:
                    sim_values.append(sim_edge[0][2])
            return sim_values
        elif element.depth <= self.reduce_depth:
            res = []
            for child in element.components:
                res += self.compute_below(child)
            return res

    def compute_above(self, element: TextElement) -> List[float]:
        if element.depth > self.reduce_depth:
            values = []
            for child in element.components:
                values += self.compute_above(child)
            element.indices[self] = self.reduce_function(values) if len(values) > 0 else ComplexityIndex.IDENTITY
        elif element.depth == self.reduce_depth:
            v = self.compute_below(element)
            if len(v) != 0:
                values = [sum(v) / len(v)]
                element.indices[self] = sum(v) / len(v)
            else:
                values = []
                element.indices[self] = ComplexityIndex.IDENTITY
        else:
            logger.error('wrong reduce depth value.')
        return values
    
    def __repr__(self):
        return self.reduce_function_abbr + self.reduce_depth_abbr + self.abbr + '_' + self.element_type.name


from os import listdir
from os.path import isfile, join

import xml.etree.ElementTree as ET
import xlrd

from rb.docs_processing.article import Article, create_article_and_add_it_to_its_authors_and_graph
from rb.docs_processing.graph import Graph
from rb.core.lang import Lang


def parse_xml_file_and_add_article_to_graph(file_path: str, graph: Graph) -> Article:
    tree = ET.parse(file_path)
    root = tree.getroot()

    abstract = root[0][0].text
    title = root[1][0].text.replace('\n', ' ')
    date = root[1][1].text
    source = root[1][2].text
    authors = []
    for element in root[1][4]:
        authors.append(element.text)
    create_article_and_add_it_to_its_authors_and_graph(title, abstract, source, authors, date, Lang.EN, graph)


def parse_santiago_xls_file_and_add_articles_to_graph(file_path: str, graph: Graph):
    workbook = xlrd.open_workbook(file_path)
    sheet = workbook.sheet_by_index(0)

    for i in range(1, 217):
        date = "01-01-" + str(int(sheet.cell_value(i, 1)))
        title = sheet.cell_value(i, 4)
        authors = sheet.cell_value(i, 2).split(";")
        abstract = sheet.cell_value(i, 12)
        source = "SANTIAGO"
        if not abstract or not title or not authors:
            continue
        create_article_and_add_it_to_its_authors_and_graph(title, abstract, source, authors, date, Lang.EN, graph)


def load_directory_xmls(directory_path: str) -> Graph:
    graph = Graph()
    for f in listdir(directory_path):
        file_path = join(directory_path, f)
        if isfile(file_path):
            print(file_path)
            parse_xml_file_and_add_article_to_graph(file_path, graph)
    return graph


def parse_xls_file_and_add_articles_to_graph(file_path: str, graph: Graph):
    workbook = xlrd.open_workbook(file_path)
    sheet = workbook.sheet_by_index(0)

    for i in range(1, sheet.nrows):
        date = "01-01-" + str(int(sheet.cell_value(i, 1)))
        title = sheet.cell_value(i, 2)
        authors = sheet.cell_value(i, 3).split(",")
        abstract = sheet.cell_value(i, 4)
        source = "NIC"
        if not abstract or not title or not authors:
            continue
        create_article_and_add_it_to_its_authors_and_graph(title, abstract, source, authors, date, Lang.EN, graph)


def load_directory_nic_xls(directory_path: str) -> Graph:
    graph = Graph([])
    for f in listdir(directory_path):
        file_path = join(directory_path, f)
        if isfile(file_path):
            print(file_path)
            parse_xls_file_and_add_articles_to_graph(file_path, graph)
    return graph


def load_directory_santiago_xls(directory_path: str) -> Graph:
    graph = Graph([])
    for f in listdir(directory_path):
        file_path = join(directory_path, f)
        if isfile(file_path):
            print(file_path)
            parse_santiago_xls_file_and_add_articles_to_graph(file_path, graph)
    return graph

# load_directory_xmls("C:\\Users\\Dragos\\Documents\\Facultate-Munca\\onlinedatasetexplorer\\AI_grub")




from typing import Callable, Dict, List, Tuple, Union

import networkx as nx
import numpy as np
from rb.cna.edge_type import EdgeType
from rb.cna.overlap_type import OverlapType
from rb.core.block import Block
from rb.core.document import Document
from rb.core.lang import Lang
from rb.core.pos import POS
from rb.core.text_element import TextElement
from rb.core.word import Word
from rb.similarity.vector_model import VectorModel


class CnaGraph:
    def __init__(self, docs: Union[Document, List[Document]], models: List[VectorModel]):
        if isinstance(docs, Document):
            docs = [docs]
        self.graph = nx.MultiDiGraph()
        for doc in docs:
            self.add_element(doc)
        self.models = models
        levels = dict()
        for n in self.graph.nodes:
            if not n.is_word():
                if n.depth in levels:
                    levels[n.depth].append(n)
                else:
                    levels[n.depth] = [n]
        
        for depth, elements in levels.items():
            self.add_lexical_links(elements, lambda w: w.pos in {POS.ADJ, POS.ADV, POS.NOUN, POS.VERB}, OverlapType.CONTENT_OVERLAP)
            self.add_lexical_links(elements, lambda w: w.pos in {POS.NOUN, POS.VERB}, OverlapType.TOPIC_OVERLAP)
            self.add_lexical_links(elements, lambda w: w.pos in {POS.NOUN, POS.PRON}, OverlapType.ARGUMENT_OVERLAP)
            self.add_semantic_links(elements)
        self.importance = self.compute_importance()
        if docs[0].lang == Lang.EN:
            self.add_coref_links()
        doc.cna_graph = self
        
    def add_element(self, element: TextElement):
        self.graph.add_node(element)
        if not element.is_sentence():
            for child in element.components:
                self.add_element(child)
                self.graph.add_edge(child, element, type=EdgeType.PART_OF)
            self.graph.add_edges_from(zip(element.components[:-1], element.components[1:]), type=EdgeType.ADJACENT)
    
    def add_semantic_links(self, elements: List[TextElement]):
        for i, a in enumerate(elements[:-1]):
            for b in elements[i+1:]:
                for model in self.models:
                    sim = model.similarity(a, b)
                    self.graph.add_edge(a, b, type=EdgeType.SEMANTIC, model=model, value=sim)
                    self.graph.add_edge(b, a, type=EdgeType.SEMANTIC, model=model, value=sim)
        
    def add_lexical_links(self, elements: List[TextElement], test: Callable[[Word], bool], link_type: OverlapType):
        for i, a in enumerate(elements[:-1]):
            for b in elements[i+1:]:
                words_a = {word.lemma for word in a.get_words() if test(word)}
                words_b = {word.lemma for word in b.get_words() if test(word)}
                weight = len(words_a & words_b) / (1e-5 + len(words_a | words_b))
                self.graph.add_edge(a, b, type=EdgeType.LEXICAL_OVERLAP, model=link_type, value=weight)
                self.graph.add_edge(b, a, type=EdgeType.LEXICAL_OVERLAP, model=link_type, value=weight)
                    
    def add_coref_links(self):
        for node in self.graph.nodes():
            if isinstance(node, Block):
                if node.has_coref:
                    for cluster in node.coref_clusters:
                        for mention in cluster.mentions:
                            if mention != cluster.main and mention.container != cluster.main.container:
                                self.graph.add_edge(mention.container, cluster.main.container, type=EdgeType.COREF)


    def compute_importance(self) -> Dict[TextElement, float]:
        similarities = [value for _, _, value in self.edges(None, edge_type=EdgeType.SEMANTIC)]
        mean = np.mean(similarities)
        stdev = np.std(similarities)
        importance = {}
        for node in self.graph.nodes:
            importance[node] = sum([value for _, _, value in self.edges(node, edge_type=EdgeType.SEMANTIC) if value > mean + stdev])
        return importance

    def edges(self, 
            node: Union[TextElement, Tuple[TextElement, TextElement]], 
            edge_type: EdgeType = None, 
            vector_model: Union[VectorModel, OverlapType] = None) -> List[Tuple[TextElement, TextElement, float]]:
        return [(a, b, data["value"] if "value" in data else 0)
            for a, b, data in self.graph.edges(node, data=True) 
            if (edge_type is None or data["type"] is edge_type) and 
               (vector_model is None or data["model"] is vector_model)
        ]



from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoNumTypeEnum(Enum):
    CARD = auto()
    ORD = auto()


class RoPOSFeatureNumType(POSFeature):


    _INSTANCE = None
    
    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.NRT,
                            values=None, pos_supported=None)
        self.values = set([ro_nr_type for _, ro_nr_type in RoNumTypeEnum.__members__.items()])
        self.pos_supported = set([POS.NUM])

    @classmethod
    def get_instance(cls) -> "RoPOSFeatureNumType":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeatureNumType()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoNumTypeEnum]:
        pattern = '(NumType=[a-zA-Z,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        nr_forms = matches[0].split('=')[1].split(',')
        values = []
        for nr_form in nr_forms:
            if nr_form == 'Card':
                values.append(RoNumTypeEnum.CARD)
            elif nr_form == 'Ord':
                values.append(RoNumTypeEnum.ORD)
        return values


from gensim.models.keyedvectors import KeyedVectors
from gensim.models import Word2Vec, LdaMulticore, LsiModel, FastText, KeyedVectors, LdaModel
from gensim.test.utils import datapath, get_tmpfile
from gensim.corpora import Dictionary, MmCorpus
import gensim
import pyLDAvis
import pyLDAvis.gensim
from rb.utils.utils import split_sentences, tokenize_docs, load_docs
from typing import Dict, List, Iterable
from os.path import dirname
import numpy as np
from itertools import islice
from rb.parser.spacy_parser import SpacyParser
from math import log
import time
import itertools
from rb.core.lang import Lang
from rb.utils.rblogger import Logger
import logging
from gensim.models.doc2vec import TaggedDocument, Doc2Vec
from os import listdir
from os.path import isdir, isfile, join

logger = Logger.get_logger()
# use this to disable loggins for gensim library
logging.getLogger("gensim").setLevel(logging.WARNING)

class Preprocess(object):
    def __init__(self, parser: SpacyParser, folder: str, 
                lang: Lang, split_sent: bool = True, only_dict_words: bool = False) -> None:
        if only_dict_words:
            self.test = lambda x: not x.is_oov
        else:
            self.test = lambda x: True
        self.lang = lang
        self.folder = folder
        self.split_sent = split_sent
        self.only_dict_words = only_dict_words
        self.parser = parser

    def __iter__(self):
        for doc in load_docs(self.folder):
            #print('processing doc {}'.format(doc))
            result = []
            doc = self.parser.preprocess(doc, self.lang)
            for tokens in self.parser.get_tokens_lemmas(self.parser.tokenize_sentences(doc), self.lang):
                if len(tokens) == 0:
                    continue
                tokens = [token.lemma_.lower() for token in tokens if not token.is_stop and token.is_alpha and self.test(token)]
                if self.split_sent:
                    yield tokens
                else:
                    result += tokens
            if not self.split_sent:
                yield result

def train_w2v(sentences: Preprocess, outputFolder: str):
    global logger
    logger.info("Starting training word2vec...")
    model = Word2Vec(size=300, window=5, min_count=5, workers=16)
    model.build_vocab(sentences=sentences)
    total_words = model.corpus_total_words  # number of words in the corpus
    total_examples = model.corpus_count # examples aka sentences
    logger.info("Words in vocabulary: {}, examples: {}".format(total_words, total_examples))

    model.train(sentences=sentences, total_words=total_words, total_examples=total_examples, epochs=5)  # train
    path = outputFolder + "/word2vec.model"
    model.save(path)
    logger.info("Model word2vec saved to {}".format(path))

# load_word2vec_format should be True if the model was saved with saved  model.wv.save_word2vec_format(path, binary=False)
# if the model was saved simply with model.save(path) load_word2vec_format should be False
def test_load_w2v(path_w2vec: str, load_word2vec_format=False):
    test_words = ["alunec", 'merg', 'ajung', 'plec', 'coleg']
    if load_word2vec_format == True:
        model = KeyedVectors.load_word2vec_format(path_w2vec, binary=False)
    else:
        model = Word2Vec.load(path_w2vec)
    for tw in test_words:
        if tw in model.wv:
            logger.info('Word vector for {} {}'.format(tw, model.wv[tw]))
    logger.info('word2vec loaded successfully')

def train_fast_text(sentences: Preprocess, outputFolder: str):
    logger.info("Starting training fast text ...")
    model = FastText(size=300, window=5, min_count=5, workers=16)
    model.build_vocab(sentences=sentences)
    total_words = model.corpus_total_words  # number of words in the corpus
    total_examples = model.corpus_count # examples aka sentences
    logger.info("Words in vocabulary: {}, examples: {}".format(total_words, total_examples))

    model.train(sentences=sentences, total_words=total_words, total_examples=total_examples, epochs=5)  # train
    path = outputFolder + "/fast_text.model"
    model.save(path)
    logger.info("Model fast text saved to {}".format(path))

def test_load_fast_text(path_fast_text: str):
    fname = get_tmpfile(path_fast_text)
    test_word = "istorie"
    model = FastText.load(fname)
    logger.info('Word vector for {} {}'.format(test_word, model.wv[test_word]))

def train_lda(docs: List, outputFolder: str):
    docs = list(docs)
    id2word = Dictionary(docs)
    id2word.filter_extremes(no_below=8, no_above=0.1, keep_n=1000000)
    logger.info('Starting trainin lda model with {} docs {} tokens'.format(id2word.num_docs, id2word.num_pos))
    corpus = [id2word.doc2bow(doc) for doc in docs]
    # saved for visualuzations
    MmCorpus.serialize(join(outputFolder, 'lda_corpus.mm'), corpus)
    logger.info('Corpus saved.')
    logger.info("Starting training lda...")
    lda = LdaMulticore(corpus, num_topics=300, id2word=id2word)
    path = join(outputFolder, "lda.model")

    #matrix = np.transpose(lda.get_topics())
    # with open(path, "wt", encoding='utf-8') as f:
    #     f.write("{} {}\n".format(np.size(matrix, 0), np.size(matrix, 1)))
    #     for idx in range(np.size(matrix, 0)):
    #         f.write(id2word[idx] + " " + " ".join([str(x) for x in matrix[idx]]) + "\n")
    lda.save(path)
    logger.info("Model lda saved to {}".format(path))

def test_load_lda(path_to_model: str):
    model = LdaModel.load(path_to_model)
    for i in range(10):
        logger.info(model.show_topic(i, 30))
    logger.info('LDA loaded successfully')

def visualize_lda(path_to_lda: str, path_to_corpus_lda: str):
    logger.info('Loadin lda model ...')
    corpus = gensim.corpora.MmCorpus(path_to_corpus_lda)
    ldaModel = gensim.models.LdaModel.load(path_to_lda)
    id2word = ldaModel.id2word
    p = pyLDAvis.gensim.prepare(ldaModel, corpus, id2word,  mds='mmds')
    pyLDAvis.save_html(p, 'lda.html')
    logger.info('LDA visualization saved in lda.html file')

def log_entropy_norm(corpus: List) -> List:
    result = []
    for doc in corpus:
        if len(doc) == 0:
            result.append(doc)
            continue
        total = sum(count for idx, count in doc)
        entropy = sum((count / total) * log(count / total) for idx, count in doc)
        entropy = 1 + (entropy / log(len(corpus)))
        result.append([(idx, log(1 + count) * entropy) for idx, count in doc])
    return result


def train_lsa(docs: Iterable, outputFolder: str):
    docs = list(docs)
    id2word = Dictionary(docs)
    id2word.filter_extremes(no_below=7, no_above=0.1, keep_n=1000000)
    corpus = [id2word.doc2bow(doc) for doc in docs]
    corpus = log_entropy_norm(corpus)
    logger.info('Lsa, docs: {}, dictionary: {}'.format(len(docs), len(id2word)))
    logger.info("Starting training lsa...")
    
    lsa = LsiModel(corpus=corpus, id2word=id2word, num_topics=300)
    path = outputFolder + "/lsa.model"
    lsa.save(outputFolder + "/lsa.bin")
    matrix = np.transpose(lsa.get_topics())
    # lsa.get_topics() shape is  (num_topics, vocabulary_size)
    # transpose to use it as word embeddings
    with open(path, "wt", encoding='utf-8') as f:
        f.write("{} {}\n".format(np.size(matrix, 0), np.size(matrix, 1)))
        for idx in range(np.size(matrix, 0)):
            f.write(id2word[idx] + " " + " ".join([str(x) for x in matrix[idx]]) + "\n")
    logger.info("Model lsa saved to {}".format(path))

# TODO, model was saved manually, load manually
def test_load_lsa(path_lsa: str):
    model = LsiModel.load(path_lsa)
    logger.info(model.projection.s)
    logger.info('LSA loaded successfully')

# parse tokens, get lemmas, only lowercase letters
def preprocess(parser: SpacyParser, folder: str, lang: Lang,
     split_sent: bool = True, only_dict_words: bool = False) -> Iterable[List]:
    if only_dict_words:
        test = lambda x: not x.is_oov
    else:
        test = lambda x: True
    for doc in load_docs(folder):
        #print('processing doc {}'.format(doc))
        result = []
        doc = parser.preprocess(doc, lang)
        for tokens in parser.get_tokens_lemmas(parser.tokenize_sentences(doc), lang):
            if len(tokens) == 0:
                continue
            tokens = [token.lemma_.lower() for token in tokens if not token.is_stop and token.is_alpha and test(token)]
            if split_sent:
                yield tokens
            else:
                result += tokens
        if not split_sent:
            yield result

if __name__ == "__main__":
    #inputFolder= "resources/corpora/FR/Le Monde"
    inputFolder = "/home/teo/projects/newsanalyser/dumped-news-sample"
    parser = SpacyParser.get_instance()
    logger.info("Loading dataset from {}".format(inputFolder))
    sentences = Preprocess(parser, inputFolder, Lang.RO, split_sent=False, only_dict_words=False)
    # #train_fast_text(sentences, inputFolder)
    #train_lsa(sentences, inputFolder)
    #train_w2v(sentences, inputFolder)
    train_lda(sentences, inputFolder)
    
    visualize_lda(path_to_lda=join(inputFolder, "lda.model"),
                  path_to_corpus_lda=join(inputFolder, "lda_corpus.mm"))

    #test_load_w2v("/home/teo/projects/news-cosmin/models/word2vec.model", load_word2vec_format=False)
    #test_load_fast_text("/home/teo/projects/readme-models/models/fasttext2/fast_text.model")
    
    #test_load_lsa("/home/teo/projects/readme-models/models/lsa/lsa.bin")
    #test_load_w2v("/home/teo/projects/readme-models/models/word2vec.model", True)

    # model = LsiModel.load(inputFolder + "/lsa.bin")
    # print(model.projection.s)
#from gensim.models.keyedvectors import KeyedVector



from typing import List

from typing import List

from rb.core.lang import Lang
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType
from rb.complexity.measure_function import MeasureFunction
from rb.cna.cna_graph import CnaGraph

# create all indices
# dependencies need to be putted in function because otherwise circular dependencies happens
def create(lang: Lang, cna_graph: CnaGraph) -> List["ComplexityIndex"]:
    from rb.complexity.cohesion.adj_cohesion import AdjCohesion
    from rb.complexity.cohesion.adj_ext_cohesion import AdjExternalCohesion
    from rb.complexity.cohesion.intra_cohesion import IntraCohesion
    from rb.complexity.cohesion.start_end_cohesion import StartEndCohesion
    from rb.complexity.cohesion.start_mid_cohesion import StartMiddleCohesion
    from rb.complexity.cohesion.mid_end_cohesion import MiddleEndCohesion
    from rb.complexity.cohesion.trans_cohesion import TransCohesion

    indices = []
    if cna_graph:
        indices.append(IntraCohesion(lang=lang, element_type=TextElementType.BLOCK, 
                                        reduce_depth=TextElementType.BLOCK.value, reduce_function=MeasureFunction.AVG,
                                        cna_graph=cna_graph))
        """ aka inter cohesion """
        indices.append(IntraCohesion(lang=lang, element_type=TextElementType.DOC, 
                                        reduce_depth=TextElementType.DOC.value, reduce_function=MeasureFunction.AVG,
                                        cna_graph=cna_graph))
        indices.append(AdjCohesion(lang=lang, element_type=TextElementType.DOC, 
                                        reduce_depth=TextElementType.DOC.value, reduce_function=MeasureFunction.AVG,
                                        cna_graph=cna_graph))
        indices.append(AdjCohesion(lang=lang, element_type=TextElementType.DOC, 
                                        reduce_depth=TextElementType.DOC.value, reduce_function=MeasureFunction.STDEV,
                                        cna_graph=cna_graph))
        indices.append(AdjCohesion(lang=lang, element_type=TextElementType.BLOCK, 
                                        reduce_depth=TextElementType.DOC.value, reduce_function=MeasureFunction.AVG,
                                        cna_graph=cna_graph))
        indices.append(AdjCohesion(lang=lang, element_type=TextElementType.BLOCK, 
                                        reduce_depth=TextElementType.DOC.value, reduce_function=MeasureFunction.STDEV,
                                        cna_graph=cna_graph))
        indices.append(AdjExternalCohesion(lang=lang, element_type=TextElementType.SENT,
                                            cna_graph=cna_graph))
        indices.append(AdjExternalCohesion(lang=lang, element_type=TextElementType.BLOCK,
                                            cna_graph=cna_graph))
        indices.append(StartEndCohesion(lang=lang,
                                        reduce_depth=None, reduce_function=None,
                                        cna_graph=cna_graph))
        indices.append(StartMiddleCohesion(lang=lang,
                                        reduce_depth=None, reduce_function=None,
                                        cna_graph=cna_graph))
        indices.append(MiddleEndCohesion(lang=lang,
                                        reduce_depth=None, reduce_function=None,
                                        cna_graph=cna_graph))
        indices.append(TransCohesion(lang=lang,
                                    reduce_depth=None, reduce_function=None,
                                    cna_graph=cna_graph))
                                
    return indices


from rb.comprehension.utils.graph.cm_node_do import CmNodeDO, CmNodeType
from rb.comprehension.utils.memory.word_activation import WordActivation
from rb.comprehension.utils.graph.cm_graph_do import CmGraphDO

from typing import Dict

ActivationMap = Dict[CmNodeDO, float]


class HistoryKeeper():

    def __init__(self):
        self.activation_history = []
        self.unique_word_list   = []

    
    def save_scores(self, activation_map: ActivationMap) -> None:
        activation_auxiliar = {}

        for key in activation_map.keys():
            activation_auxiliar[key.word] = WordActivation(activation_map[key], key.active)
        
        self.activation_history.append(activation_auxiliar)

    
    def save_nodes(self, graph: CmGraphDO) -> None:
        for node in graph.node_list:
            self.add_node_if_not_existing(node)

    
    def add_node_if_not_existing(self, node: CmNodeDO) -> None:
        exists = False
        for index, current_node in enumerate(self.unique_word_list):
            if current_node == node:
                if current_node.node_type != CmNodeType.TextBased and node.node_type == CmNodeType.TextBased:
                    self.unique_word_list[index] = node
                exists = True
                break

        if not exists:
            self.unique_word_list.append(node)



#!/usr/bin/env python
# coding: utf-8

import re
import json
import nltk
import spacy
import sys
from rb.parser.spacy_parser import SpacyParser
from rb.parser.spacy_parser import convertToPenn
from rb.core.lang import Lang
from nltk.stem.snowball import SnowballStemmer
from rb.processings.ro_corrections.get_pos_properties import *
from rb.processings.ro_corrections.get_exceptions import *
from rb.processings.ro_corrections.utils import *
from rb.utils.rblogger import Logger
import os

from nltk.stem.snowball import SnowballStemmer

logger = Logger.get_logger()

stemmer = SnowballStemmer("romanian")
rom_spacy = SpacyParser.get_instance().get_model(Lang.RO)

window_size = 8


with open(os.path.join(os.path.dirname(__file__), './lexicons.json'), 'rt') as fin:
    lexicons = json.load(fin)

with open(os.path.join(os.path.dirname(__file__), './exceptions.json'), 'rt') as fin:
    exceptions = json.load(fin)

def remove_punctuation(doc):
    punctuation_list = [".", ",", "!", ":", "?", ";"]
    result = []
    
    for el in doc:
        # if el.text not in punctuation_list:
        result.append(el)
    return result

def convert_char_ind_to_word_ind(text, index):
    if index == 0:
        return {'paragraph': 0, 'index': 0}
    if '\n' not in text[:index]:
        prev_endl_index = 0
    else:
        prev_endl_index = index - list(reversed(text[:index])).index('\n') - 1
    if prev_endl_index == 0:
        paragraph_no = 0
    elif '\n' not in text[:prev_endl_index]:
        paragraph_no = 1
    else:
        paragraph_no = len(text[:prev_endl_index].split('\n'))

    index = index - prev_endl_index - 1
    paragraph = text[(prev_endl_index + 1):].split('\n')[0]
    return {'paragraph': paragraph_no, 'index': len(rom_spacy(paragraph[:index]))}

def is_exception(mistake, case):
    mistake = " ".join(mistake)
    for exception in exceptions[case]:
        if mistake == exception.lower():
            return True
    return False

#for cacophonies and "ca_si" returns a list with (mistake, index)
def build_response(mistakes_list, sentence, case):
    items = []
    if case == "Cacofonie" or case == "Si_parazitar":
        item  = {
            'mistake': case,
            'index': mistake_list
        }
        items.append(item)
        return items
    if case == "Conjunctie coordonatoare" or case == "Verbul 'a voi'" or case == "Totusi" or case == "Adverbe la inceput" or case == "Adverbe intercalate":
        for mistake in mistakes_list:
            for match in re.finditer(mistake, sentence):
                index = match.start()
                test_index = index
                if case == 'Adverbe intercalate':
                    test_index += 2
                par_word = convert_char_ind_to_word_ind(sentence, test_index)
                paragraph_index = par_word['paragraph']
                word_index = par_word['index']
                
                item  = {
                    'mistake': mistake,
                    'paragraph': paragraph_index,
                    'word_index': [word_index], 
                    'start': match.start(),
                    'end': match.end()
                }
                if item not in items:
                    items.append(item)
        return items
    if case == "Conjunctie adversativa":
        for mistake in mistakes_list:
            for match in re.finditer(mistake, sentence):
                index = match.start() + 1
                par_word = convert_char_ind_to_word_ind(sentence, index)
                paragraph_index = par_word['paragraph']
                word_index = par_word['index']
                
                item  = {
                    'mistake': mistake,
                    'paragraph': paragraph_index,
                    'word_index': [word_index], 
                    'start': match.start() + 1,
                    'end': match.end()
                }
                if item not in items:
                    items.append(item)
        return items
    if case == "Comparativ de superioritate":
        for mistake in mistakes_list:
            for match in re.finditer(mistake, sentence):
                index = match.end() - 2
                par_word = convert_char_ind_to_word_ind(sentence, index)
                paragraph_index = par_word['paragraph']
                word_index = par_word['index']
                
                item  = {
                    'mistake': mistake,
                    'paragraph': paragraph_index,
                    'word_index': [word_index], 
                    'start': match.end() - 2, 
                    'end': match.end()
                }
                if item not in items:
                    items.append(item)
        return items 

def check_cacophony(par_index, doc):
    cac_list = [['la', 'la'], ['sa', 'sa'],
            ['ca', 'ca'], ['ca', 'că'],
            ['ca', 'ce'], ['ca', 'ci'],
            ['că', 'ca'], ['cu', 'co'],
            ['că', 'cu'], ['că', 'când'],
            ['ca', 'când'], ['că', 'câ']]
    cacophony_list = []
    for i, token in enumerate(doc):
        if i == len(doc) - 2:   continue
        for c_pair in cac_list:
            if doc[i].text == c_pair[0] and doc[i + 1].text == c_pair[1]:
                cacophony_list.append({
                    'mistake': "Cacofonie",
                    'index': [ [par_index, i], [par_index, i + 1] ]
                })
    return cacophony_list

# def check_si_parazitar(sentence):
#     regex = '(\w*ca)\s(şi\w*)'
#     p = re.compile(regex, re.IGNORECASE)
#     return build_response(p.findall(sentence), sentence, "Si_parazitar")

#group repetition in dictionary to be solve separated
def build_repetitions_dictionary(repetitions):
    d = {}
    for item in repetitions:
        if item[2] not in d:
            d[item[2]] = []
        d[item[2]].append((item[0], item[1]))
    return d


#using a window check if there are words with the same stem in window. Eliminate stop_words
#create a list with (word, index, stem)
#returns a dictionary with grouped repetitions
def check_repetitions(par_index, doc):
    
    repetitions = []
    index = 0
    
    while index == 0 or len(doc) - index >= window_size:
        repetitions_dictionary = {}
        #lemmatization
        for i in range(index, min(index + window_size, len(doc))):
            word = doc[i]
            num_endlines = len(list(filter(lambda x: len(x.text.strip()) == 0, doc[:i])))
            if word.is_stop != True:
                lemma = word.lemma_
                if lemma in repetitions_dictionary:
                    repetitions_dictionary[lemma].append(i)
                else:
                    repetitions_dictionary[lemma] = [i]
        for k, v in repetitions_dictionary.items():
            if len(v) >= 2 and doc[v[0]].is_punct == False and doc[v[0]].is_stop == False:
                el = [ [par_index, value] for value in v]
                if el not in repetitions:
                    repetitions.append(el)
        index += 1
    items = []
    for rep in repetitions:
        items.append({'mistake': "Repetiție",
                    'index': rep})
    return items


def are_synonyms(word1, word2):
    return ((len(set(lexicons[word1]) & set([word2])) >= 1) or (len(set(lexicons[word2]) & set([word1])) >= 1))

def get_synonym_pairs(window):
    pairs = []
    
    for i in range(len(window)):
        word_i = window[i].lemma_
        if window[i].is_stop or word_i not in lexicons:
            continue
        for j in range(i + 1, len(window)):
            word_j = window[j].lemma_
            if window[j].is_stop or word_j not in lexicons:
                continue
            if word_i != word_j and are_synonyms(word_i, word_j):
                pairs.append((i, j))
    
    return pairs

def get_synonym_clusters(synonyms):
    clusters = {}
    cluster_mappings = {}
    for i, j in synonyms:
        if i in clusters:
            clusters[i].append(j)
            cluster_mappings[j] = i
        elif i in cluster_mappings:
            if j not in clusters[cluster_mappings[i]]:
                clusters[cluster_mappings[i]].append(j)
                cluster_mappings[j] = cluster_mappings[i]
        else:
            clusters[i] = [i, j]
            cluster_mappings[i] = i
            cluster_mappings[j] = i
    return clusters

#check if a synonym is used in window
def check_synonyms_repetitions(sentence):

    doc = rom_spacy.parser(rom_spacy(sentence))
    sentence_tokens = remove_punctuation(list(doc))

    repetitions = []
    check = set()
    index = 0
    while index == 0 or len(sentence_tokens) - index >= window_size:
        repetitions_dictionary = {}
        window = [e for e in sentence_tokens[index: index + window_size] if len(e.text.strip()) != 0]
        num_endlines = len([e for e in sentence_tokens[index: index + window_size] if len(e.text.strip()) == 0])
        synonym_pairs = get_synonym_pairs(window)
        clusters = get_synonym_clusters(synonym_pairs)
        for cluster in clusters:
            common_key = window[clusters[cluster][0]].text
            for word_index in clusters[cluster]:
                actual_index = index + word_index
                item = (window[word_index].text, actual_index + 1 + num_endlines, common_key)
                if item not in check:
                    check.add(item)
                    repetitions.append((window[word_index].text, actual_index + 1, common_key))

        index += 1
    repetitions = build_repetitions_dictionary(repetitions)
    return convert_repetitions_dictionary(sentence, repetitions)

#Dupa totusi nu se pune virgula.
def check_totusi(par_index, doc):
    mistake_list = []
    for i, token in enumerate(doc):
        if i == 0:   continue
        if doc[i].text.lower() == 'totuși' and doc[i - 1].text == ',':
            mistake_list.append({
                'mistake': "Virgulă înainte de totuși",
                'index': [ [par_index, i - 1], [par_index, i] ]
            })
    return mistake_list

#Atunci când părțile de propoziție sau propozițiile se află într-un raport de coordonare, introduse prin conjuncțiile și, sau, ori, totuși nu se despart prin virgulă.
def check_coordinative_conjunctions(sentence):
    coordinative_conjunctions_list = ['(,\s*și)','[^\w](și\s*,)', '(,\s*sau)','(sau\s*,)', '(,\s*ori)','(ori\s*,)']
    mistakes = []
    for regex in coordinative_conjunctions_list:
        p = re.compile(regex, re.IGNORECASE)
        mistakes += p.findall(sentence)
    return build_response(mistakes, sentence, "Conjunctie coordonatoare")

def check_adversative_conjunctions(sentence):
    #nici, dar, iar, însă, ci, ci și, dar și, precum și, deci, prin urmare, așadar, în concluzie.
    adversative_conjunctions = ['[a-z]\s+dar\s', '[a-z]\s+nici\s', '[a-z]\s+iar\s', '[a-z]\s+însă\s', '[a-z]\s+ci\s','[a-z]\s+ci\s+și\s',
                               '[a-z]\s+dar\s+și\s]', '[a-z]\s+precum\s+și\s', '[a-z]\s+deci\s', '[a-z]\s+prin\s+urmare\s',
                               '[a-z]\s+așadar\s', '[a-z]\s+în\s+concluzie\s']
    mistakes = []
    for regex in adversative_conjunctions:
        p = re.compile(regex, re.IGNORECASE)
        mistakes += p.findall(sentence)
    return build_response(mistakes, sentence, "Conjunctie adversativa")


def check_voi_verb(par_index, doc):
    verb_list = ['vroiam', 'vroiai', 'vroia', 'vroiau', 'vroiaţi']
    mistake_list = []
    for i, token in enumerate(doc):
        if i == len(doc) - 2:   continue
        for vb in verb_list:
            if doc[i].text == vb:
                mistake_list.append({
                    'mistake': "Forme verbului",
                    'index': [ [par_index, i] ]
                })
    return mistake_list

def check_comparative(sentence):
    regex_list = ['(mai\s\w+\sde)', '(mai\s\w+\sca)']
    mistakes = []
    for regex in regex_list:
        p = re.compile(regex, re.IGNORECASE)
        mistakes += p.findall(sentence)
    return build_response(mistakes, sentence, "Comparativ de superioritate")

def check_adverbs_at_the_beginning(par_index, doc):

    adverb_list = ['Totuși', 'Așadar', 'Prin urmare', 'Deci']
    mistake_list = []

    for i, token in enumerate(doc):
        if i == len(doc) - 2:   continue
        for vb in adverb_list:
            if doc[i].text == vb:
                mistake_list.append({
                    'mistake': "Adevrb nepotrivit la început de frază",
                    'index': [ [par_index, i] ]
                })
    return mistake_list

def check_adverbs_in_middle(sentence):
    adverb_list = ['(\w\s+desigur)','(desigur\s+\w)', '(\w\s+firește)', '(firește\s+\w)','(\w\s+așadar)','(așadar\s+\w)',
                   '(\w\s+bineînțeles)', '(bineînțeles\s+\w)','(\w\s+în concluzie)', '(în concluzie\s+\w)', 
                   '(\w\s+în realitate)', '(în realitate\s+\w)', '(\w\s+de exemplu)', '(de exemplu\s+\w)']
    mistakes = []
    for regex in adverb_list:
        p = re.compile(regex)
        mistakes += p.findall(sentence)
    return build_response(mistakes, sentence, "Adverbe intercalate")

def check_verbal_predicate(par_index, S, P, spn, corrections):
    try:
        if P.tag_[:3] == "Vmi":
            spv = get_verb(P.tag_)
            if is_collective_subject(S, spv, exceptions):
                return None

            if spn[:2] == spv:
                return None
            if spn[0] == spv[0] and spv[1] == "":
                return None
            if spn[1] == spv[1] and spv[0] == "-":
                return None
            
            person, number, gender = get_person_number_gender(spn)
            mistake = {
                'mistake': 'Acordul verbului cu subiectul',
                'index': [[par_index, S.i], [par_index, P.i]]
            }
            corrections.append(mistake)
            
        elif P.tag_[:3] == "Vmp" or P.tag_[:4] == "Vmnp":
            for child in P.children:
                if child.dep_ == "aux":
                    if child.tag_ == "Vanp":
                        return None

                    aux = get_verb(child.tag_)
                    if is_collective_subject(S, aux, exceptions):
                        return None

                    if aux == spn[:2]:
                        return None
                    if aux[0] == spn[0] and aux[1] == "":
                        return None

                    mistake = {
                        'mistake': 'Acordul verbului cu subiectul',
                        'index': [[par_index, S.i], [par_index, child.i]]
                    }
                    corrections.append(mistake)
    except:
        return None

def check_copulative_verb(par_index, S, P, spn, corrections):
    #check copulativ verb with subject
    has_cop = False
    for child in P.children:
        if child.dep_ == "cop":
            if child.tag_[:3] == "Vmi" or child.tag_[:3] == "Vai":
                spv = get_verb(child.tag_)

                if is_collective_subject(S, spv, exceptions):
                    return None
                
                if spn[:2] == spv:
                    return None
                if spn[0] == spv[0] and spv[1] == "":
                    return None
                if spn[1] == spv[1] and spv[0] == "-":
                    return None
                # predicat nominal: verb copulativ + nume predicativ, punele pe ambele in index
                mistake = {
                    'mistake': 'Acordul verbului cu subiectul',
                    'index': [[par_index, S.i], [par_index, P.i], [par_index, child.i]]
                }
                corrections.append(mistake)
            else:
                has_cop = True
    if has_cop == True:
        for child1 in P.children:
            if child1.dep_ == "aux":
                if child1.tag_ == "Vanp":
                    return None
                aux = get_verb(child1.tag_)
                if is_collective_subject(S, aux, exceptions):
                    return None

                if aux == spn[:2]:
                    return None
                if aux[0] == spn[0] and aux[1] == "":
                    return None

                mistake = {
                    'mistake': 'Acordul verbului cu subiectul',
                    'index': [[par_index, S.i], [par_index, child.i]]
                }
                corrections.append(mistake)

def check_predicative_name(par_index, S, NP, spn, corrections):
    if is_job_noun(spn, NP, exceptions):
        return None

    if NP.tag_[:3] == "Rgp":
        return None

    np = ""
    #adjectiv
    if NP.tag_[:2] == "Af":
        np = get_adjective(NP.tag_)
    #noun
    elif NP.tag_[:2] == "Nc":
       np = get_noun(NP.tag_)

    #ordinal numeral or cardinal numeral
    elif NP.tag_[:2] == "Mo" or NP.tag_[:2] == "Mc":
       np = get_numeral(NP.tag_)

    if np[1] == spn[1] and np[2] == spn[2]:
       return None
    #Subject is a personal pronoun without gender
    if np[1] == spn[1] and spn[2] == "-":
       return None

    mistake = {
        'mistake': 'Acordul numelui predicativ cu subiectul',
        'index': [[par_index, S.i], [par_index, NP.i]]
    }
    corrections.append(mistake)


def check_subject_and_predicate_relation(par_index, doc):
    corrections = []
    predicates = []

    for token in doc:
        if token.dep_ == "nsubj":
            S = token
            P = token.head
            if P not in predicates:
                predicates.append(P)
            else:
                continue

            if is_multiple_subject_linked_to_subject(S):
                spn = get_multiple_subject(S, "conj", exceptions)
            elif is_multiple_subject_linked_to_predicate(P):
                spn = get_multiple_subject(P, "nsubj", exceptions)
            else:
                if S.tag_[0] == "P":
                    spn = get_pronoun(S.tag_)
                else:
                    spn = get_noun(S.tag_)

            multiple_predicates = get_multiple_predicates(P)

            #all parts from multiple_predicates shoud be in the same relation with the subject
            for single_predicate in multiple_predicates:
                if single_predicate.tag_[0] == "V":
                    check_verbal_predicate(par_index, S, single_predicate, spn, corrections)
                else:
                    check_copulative_verb(par_index, S, single_predicate, spn, corrections)
                    check_predicative_name(par_index, S, single_predicate, spn, corrections)
    return corrections

def check_noun_and_adjective_relation(par_index, doc):
    corrections = []
    for token in doc:
        if token.tag_[:2] == "Af":
            (is_linked, noun) = is_in_adjective_noun_relation(token)
            if is_linked:
                A = token
                N = noun
                spa = get_adjective(A.tag_)
                spn = get_noun(N.tag_)
                if spa[1] == spn[1] and spa[2] == spn[2]:
                    continue
                else:
                    mistake = {
                        'mistake': 'Acord substantiv - adjectiv',
                        'index': [[par_index, A.i], [par_index, N.i]]
                    }
                    corrections.append(mistake)
    return corrections

def check_noun_and_numeral_relation(par_index, doc):
    corrections = []
    for token in doc:
        if token.dep_ == "nummod":
            if token.tag_[:2] == "Mc" or token.tag_[:2] == "Mo":
                Nr = token
                N = token.head
                spnr = get_numeral(Nr.tag_)
                spn = get_noun(N.tag_)
                if spnr[1] == spn[1] and spnr[2] == spn[2]:
                    continue
                if spnr[2] == "-" and spnr[1] == spn[1]:
                    continue
                else:
                    #it is not correct
                    mistake = {
                        'mistake': 'Acord substantiv - numeral',
                        'index': [[par_index, Nr.i], [par_index, N.i]]
                    }
                    corrections.append(mistake)
    return corrections

def check_noun_and_unstated_article_relation(par_index, doc):
    corrections = []
    for token in doc:
        if token.dep_ == "det":
            if token.tag_[:2] == "Ti":
                UA = token
                N = token.head
                spar = get_unstated_article(UA.tag_)
                spn = get_noun(N.tag_)
                if spar[1] == spn[1] and spar[2] == spn[2]:
                    continue
                if spar[2] == "-" and spar[1] == spn[1]:
                    continue
                else:
                    #it is not correct

                    mistake = {
                        'mistake': "Acord substantiv - acord nehotarat",
                        'index': [[par_index, token.i], [par_index, N.i]]
                    }
                    corrections.append(mistake)
            elif token.tag_[:3] == "Di3":
                UA = token.head
                N = token.head
                spn = get_noun(N.tag_)
                if spn[2] == "p":
                    continue
                else:
                    #it is not correct
                    mistake = {
                        'mistake': "Acord substantiv - acord nehotarat",
                        'index': [[par_index, token.i], [par_index, N.i]]
                    }
                    corrections.append(mistake)
    return corrections


def create_output(category, message, paragraph_index, word_index):
    output = {}
    output["title"] = category
    output["message"] = message
    output["paragraph_index"] = paragraph_index
    output["word_index"] = word_index
    return output

def get_mistakes(mistakes):
    message = []
    for item in mistakes:
        message.append("(" + str(item[0]) + ")")
    return ", ".join(message)

def create_output_list(result, case, default_suggestion = None):
    output_list = []
    if case == "Cacofonie":
        for item in result:
            message = "Reformulare. Cacofonie."
            output_list.append(create_output(case, message, paragraph_index, item["word_index"][0]))
            output_list.append(create_output(case, message, paragraph_index, item["word_index"][1]))
        return output_list
    if case == "şi parazitar":
        for item in result: 
            message = "Expresia (" + item["mistake"] + ") Şi parazitar (apariția lui nu este permisă nici pentru evitarea cacofoniei). Reformulare folosind una dintre expresiile: 'în calitate de' sau 'drept'"
            paragraph_index = item["paragraph"]
            output_list.append(create_output(case, message, paragraph_index, item["word_index"][0]))
            output_list.append(create_output(case, message, paragraph_index, item["word_index"][1]))
        return output_list
    if case == "Repetiţie" or case == "Repetiţie sinonime":
        for mistake_list in result:
            mistake_output_list = []
            for item in mistake_list:
                try:
                    if len(item['mistake']) >= 1 and (str(item['mistake'][0][0]) == '.' or str(item['mistake'][0][0]) == ','):
                        continue
                except:
                    pass
                message = "Reformulare. Repetiţie deranjantă pentru cuvântul/cuvintele: " + get_mistakes(item["mistake"])
                if case == "Repetitie sinonime":
                    message += " Cuvintele sunt sinonime."
                paragraph_index = item["paragraph"]
                word_index = item["word_index"]
                mistake_output_list.append(create_output(case, message, paragraph_index, word_index))
                # print('len mistake output list', len(mistake_output_list))
                # print(mistake_output_list)
            output_list.append(mistake_output_list)
        return output_list
    if case == "Conjuncţie coordonatoare":
        for item in result:
            message = "Atunci când părțile de propoziție sau propozițiile se află într-un raport de coordonare, introduse prin conjuncțiile și, sau, ori, nu se despart prin virgulă. Varianta corectă: " + default_suggestion
            paragraph_index = item["paragraph"]
            word_index = item["word_index"]
            output_list.append(create_output(case, message, paragraph_index, word_index))
        return output_list
    if case == "Conjuncţie adversativă":
        for item in result:
            message = "Virgula este obligatorie în fața conjuncțiilor/ locuțiunilor coordonatoare adversative. Varianta corectă: " + default_suggestion
            paragraph_index = item["paragraph"]
            word_index = item["word_index"]
            output_list.append(create_output(case, message, paragraph_index, word_index))
        return output_list
    if case == "Verbul 'a voi'":
        for item in result:
            message = "Forma corectă de imperfect a verbului 'a voi' este 'voiam' la persoana I etc. Varianta corectă: " + default_suggestion
            paragraph_index = item["paragraph"]
            word_index = item["word_index"]
            output_list.append(create_output(case, message, paragraph_index, word_index))
        return output_list
    if case == "Comparativ de superioritate":
        for item in result:
            message = "Comparativul de superioritate se exprimă cu ajutorul lui „decât”. Varianta corectă: " + default_suggestion
            paragraph_index = item["paragraph"]
            word_index = item["word_index"]
            output_list.append(create_output(case, message, paragraph_index, word_index))
        return output_list
    if case == "Totuşi":
        for item in result:
            message = "După 'totuși' nu se pune virgulă, dacă acesta este în interiorul propoziției. Varianta corectă: " + default_suggestion
            paragraph_index = item["paragraph"]
            word_index = item["word_index"]
            output_list.append(create_output(case, message, paragraph_index, word_index))
        return output_list
    if case == "Adverbe la începutul propoziţiei":
        for item in result:
            message = "La începutul propoziției, se pune virgulă după termenii: 'totuşi', 'aşadar', 'prin urmare' şi 'deci'. Varianta corectă: " + default_suggestion
            paragraph_index = item["paragraph"]
            word_index = item["word_index"]
            output_list.append(create_output(case, message, paragraph_index, word_index))
        return output_list
    if case == "Adverbe intercalate":
        for item in result:
            message = "Intercalarea de adverbe, precum 'desigur', 'fireşte', 'aşadar', 'bineînţeles', 'în concluzie', 'în realitate', 'de exemplu' impune punerea virgulei. Varianta corectă: " + default_suggestion
            paragraph_index = item["paragraph"]
            word_index = item["word_index"]
            output_list.append(create_output(case, message, paragraph_index, word_index))
        return output_list


def sort_by_start_index(result):
    return sorted(result, key = lambda x: x['start'])


def corrects_coordonative_error(result, sentence):
    result = sort_by_start_index(result)
    
    sentence = list(sentence)
    for item in result:
        start = item["start"]
        end = item["end"]
        if sentence[start] == ",":
            sentence[start] = " "
        if sentence[end - 1] == ",":
            sentence[end - 1] = " "
    sentence = "".join(sentence)
    p = re.compile('\s\s+')
    return p.sub(' ', sentence)


def corrects_adversative_error(result, sentence):
    result = sort_by_start_index(result)
    
    sentence = list(sentence)
    for item in result:
        start = item["start"]
        sentence[start] = ", "
    sentence = "".join(sentence)
    p = re.compile('\s\s+')
    return p.sub(' ', sentence)


def corrects_voi_verb(result, sentence):
    result = sort_by_start_index(result)
    
    sentence = list(sentence)
    index = 0
    for item in result:
        del sentence[item["start"] + 1 - index]
        index += 1
    return "".join(sentence)


def corrects_comparative(result, sentence):
    result = sort_by_start_index(result)
    
    index = 0
    for item in result:
        sentence = sentence[0:item["start"] + index * 3] + "decât" + sentence[item["end"] + index * 3:]
        index += 1
    return sentence

def corrects_adverbs_at_the_beginning(result, sentence):
    result = sort_by_start_index(result)
    
    sentence = list(sentence)
    for item in result:
        end = item["end"]
        sentence[end - 1] = ", "
    sentence = "".join(sentence)
    p = re.compile('\s\s+')
    return p.sub(' ', sentence)



def corrects_adverbs_in_middle(result, sentence):
    result = sort_by_start_index(result)
    
    p = re.compile('\s\s+')
    sentence = list(sentence)
    for item in result:
        start = item["start"]
        end = item["end"]
        if sentence[start + 1] == " ":
            sentence[start + 1] = ", "
        if sentence[end - 2] == " ":
            sentence[end - 2] = ", "
    
    sentence = "".join(sentence)
    return p.sub(' ', sentence)
            


def split_text(sentence):
    split_text = []
    for sentence in sentence.split('\n'):
        words = [token.text for token in rom_spacy(sentence)]
        split_text.append([words])
    return split_text

def get_paragraph_offset(pid, text):
    offset = 0
    for par in text[:pid]:
        for word in par[0]:
            offset += len(word)
        offset += len(par[0]) - 1
    return offset

def get_word_offset(wid, paragraph, end = False):
    offset = 0
    for word in paragraph[0][:wid]:
        offset += len(word) + 1
        if word == '-':
            offset -= 2
    if end:
        offset += len(paragraph[0][wid]) - 1
    return offset


def fix_dashes(output):
    word_breaks = []
    new_text = []
    for paragraph in output['split_text']:
        par_breaks = []
        new_paragraph = [[]]
        word_breaks.append(par_breaks)
        new_text.append(new_paragraph)

        i = 0
        while i < len(paragraph[0]):
            if paragraph[0][i] == '-':
                par_breaks.append(i)
                new_paragraph[0][-1] += paragraph[0][i] + paragraph[0][i + 1]
                i += 2
            else:
                new_paragraph[0].append(paragraph[0][i])
                i += 1
    
    output['split_text'] = new_text
    for error_l in output['correction']:
        if isinstance(error_l, list):
            for error in error_l:
                paragraph = error['paragraph_index']
                breaks = word_breaks[paragraph]
                if len(breaks) == 0:
                    continue
                
                count = 0
                for i in range(len(error['word_index'])):
                    offset = 0
                    for e in breaks:
                        if error['word_index'][i] > e:
                            offset -= 2
                    error['word_index'][i] += offset
        else:
            error = error_l
            paragraph = error['paragraph_index']
            breaks = word_breaks[paragraph]
            if len(breaks) == 0:
                continue
            
            count = 0
            for i in range(len(error['word_index'])):
                offset = 0
                for e in breaks:
                    if error['word_index'][i] > e:
                        offset -= 2
                error['word_index'][i] += offset


def fix_punctuation(output):
    import string
    word_breaks = []
    new_text = []
    for paragraph in output['split_text']:
        par_breaks = []
        new_paragraph = [[]]
        word_breaks.append(par_breaks)
        new_text.append(new_paragraph)

        i = 0
        while i < len(paragraph[0]):
            if paragraph[0][i] != '-' and paragraph[0][i] in ".,;:!?":
                par_breaks.append(i)
                new_paragraph[0][-1] += paragraph[0][i]
                i += 1
            else:
                new_paragraph[0].append(paragraph[0][i])
                i += 1
    
    output['split_text'] = new_text
    for error_l in output['correction']:
        if isinstance(error_l, list):
            for error in error_l:
                paragraph = error['paragraph_index']
                breaks = word_breaks[paragraph]
                if len(breaks) == 0:
                    continue
                
                count = 0
                for i in range(len(error['word_index'])):
                    offset = 0
                    for e in breaks:
                        if error['word_index'][i] > e:
                            offset -= 1
                    error['word_index'][i] += offset
        else:
            error = error_l
            paragraph = error['paragraph_index']
            breaks = word_breaks[paragraph]
            if len(breaks) == 0:
                continue
            
            count = 0
            for i in range(len(error['word_index'])):
                offset = 0
                for e in breaks:
                    if error['word_index'][i] > e:
                        offset -= 1
                error['word_index'][i] += offset


def change_format(par_index, output):
    for error_l in output["correction"]:
        if isinstance(error_l, list):
            for error in error_l:
                if isinstance(error["word_index"], int):
                    error["word_index"] = [error["word_index"]]
        else:
            error = error_l
            if isinstance(error["word_index"], int):
                error["word_index"] = [error["word_index"]]
    
    fix_dashes(output)
    fix_punctuation(output)
    mistakes = []
    for error_l in output["correction"]:
        if isinstance(error_l, list):
            for error in error_l:
                par_offset = get_paragraph_offset(error["paragraph_index"], output["split_text"])
                word_offset_start = get_word_offset(error["word_index"][0], output["split_text"][error["paragraph_index"]])
                word_offset_end = get_word_offset(error["word_index"][-1], output["split_text"][error["paragraph_index"]], True)
                error["correction_index"] = [par_offset + word_offset_start, par_offset + word_offset_end]
        else:
            error = error_l
            par_offset = get_paragraph_offset(error["paragraph_index"], output["split_text"])
            word_offset_start = get_word_offset(error["word_index"][0], output["split_text"][error["paragraph_index"]])
            word_offset_end = get_word_offset(error["word_index"][-1], output["split_text"][error["paragraph_index"]], True)
            error["correction_index"] = [par_offset + word_offset_start, par_offset + word_offset_end]
    return output


def simplify_format(output):
    
    new_corrections = []

    for error_l in output['correction']:
        if isinstance(error_l, list):
            if len(error_l) > 0:
                refact_cor = {}
                correction_index_list = []
                refact_cor['message'] = error_l[0]['message']
                refact_cor['title'] = error_l[0]['title']
                for error in error_l:
                    correction_index_list.append(error['correction_index'])
                refact_cor['correction_index'] = correction_index_list
                new_corrections.append(refact_cor)
        else:
            error_l['correction_index'] = [error_l['correction_index']]
            if 'paragraph_index' in error_l:
                del error_l['paragraph_index']
            if 'word_index' in error_l:
                del error_l['word_index']
            new_corrections.append(error_l)
    res = {}
    res['corrections'] = new_corrections
    res['split_text'] = output['split_text']
    return res

def identify_mistake(par_index, sentence, spellchecking=False):

    doc = rom_spacy(sentence)
    output_list = []

    if spellchecking == True:
        spellcheck_list = []
        import hunspell as hs
        ro_hunspell = hs.HunSpell('/usr/share/hunspell/ro_RO.dic', '/usr/share/hunspell/ro_RO.aff')
        for i, token in enumerate(doc):
            if token.is_punct == False and ro_hunspell.spell(token.text) == False:
                spellcheck_list.append({
                    'mistake': "Ortografie",
                    'index': [ [par_index, i], [par_index, i + 1] ],
                    "suggestions": ro_hunspell.suggest(token.text)
                })

    output_list += spellcheck_list
    # print(dir(doc[0]))
    # for i, token in enumerate(doc):
    #     print('index', token.text, token.pos_, token.is_stop, token.is_punct, token.i)

    correct = True
    #Step1
    try:
        result = check_cacophony(par_index, doc)
        output_list += result
    except:
        logger.warning('Cachophony check failed')
        
    
    #Step2 - ca si
    # result = check_si_parazitar(par_index, sentence) - does not work all the time
    # output_list += result
        
    #Step3 - Repetitions
    try:
        result = check_repetitions(par_index, doc)
        output_list += result
    except:
        logger.warning('Repetitions check failed')
        
    #Step4 - Conjunctii coordonatoare - does not work all the time
    # result = check_coordinative_conjunctions(sentence) 
    # if result != []:
    #     correct = False
    #     output_list += create_output_list(result, "Conjuncţie coordonatoare", corrects_coordonative_error(result, sentence))

    #Step5 - Conjunctii adversative
    # result = check_adversative_conjunctions(sentence)
    # if result != []:
    #     correct = False
    #     output_list += create_output_list(result, "Conjuncţie adversativă", corrects_adversative_error(result, sentence))

    #Step6 - Repetitie sinonime
    # result = check_synonyms_repetitions(sentence.lower())
    # if result != {}:
    #     correct = False
    #     output_list += create_output_list(result, "Repetiţie sinonime")
    
    #Step7 - Voi
    try:
        result = check_voi_verb(par_index, doc)
        output_list += result
    except:
        logger.warning('Voi verb check failed')

        
    #Step8 - Comparativ de superioritate - does not work all the time
    # result = check_comparative(sentence)
    # if result != []:
    #     correct = False
    #     output_list += create_output_list(result, "Comparativ de superioritate", corrects_comparative(result, sentence))

    #Step9 - Totusi
    try:
        result = check_totusi(par_index, doc)
        output_list += result
    except:
        logger.warning('Totusi check failed')

    try:
        output_list += check_adverbs_at_the_beginning(par_index, doc) 
    except:
         logger.warning('Adverbs at the beginning check failed')
    #Step11 - Adverbe intercalate
    # result = check_adverbs_in_middle(sentence)
    # if result != []:
    #     correct = False
    #     output_list += create_output_list(result, "Adverbe intercalate", corrects_adverbs_in_middle(result, sentence))

    #Step12 - Subject and predicate
    try:
        output_list += check_subject_and_predicate_relation(par_index, doc)
    except:
        logger.warning('Subject and predicate relation check failed')
    # #Step13 - Noun and adjective relation
    try:
        output_list += check_noun_and_adjective_relation(par_index, doc)
    except:
        logger.warning('Noun and adjective relation check failed')

    # #Step14 - Noun and numeral relation
    try:
        output_list += check_noun_and_numeral_relation(par_index, doc)
    except:
        logger.warning('Noun and numeral relation check failed')
    # #Step15 - Noun and unstated article relation
    try:
        output_list += check_noun_and_unstated_article_relation(par_index, doc)
    except:
        logger.warning('Noun and unstated article relation check failed')
    return output_list, [token.text for token in doc]


def correct_text_ro(text, spellchecking=False):
    p_newline_after = re.compile('\n\s+')
    p_newline_before = re.compile('\s+\n')
    text = p_newline_after.sub('\n', text)
    text = p_newline_before.sub('\n', text)
    paragraphs = text.split('\n')
    paragraphs = [p for p in paragraphs if len(p) > 0]

    output = {'split_text': [], 'corrections': []} 
    for i, paragraph in enumerate(paragraphs):
        mistakes, text_splitted = identify_mistake(i, paragraph, spellchecking=spellchecking)
        output['split_text'].append(text_splitted)
        output['corrections'] += mistakes
    return output

if __name__ == "__main__":
    txt = "Fiind protejate de stratul de gheaţă, apele mai adânci nu îngheaţă până la fund, ci au, sub stratul de gheaţă, temperatura de 4 grade la care viaţa poate continua"
    correct_text_ro(txt)
    with open('log.log', 'wt', encoding='utf-8') as f:
        f.write(jsonify(correct_text_ro(txt)))


import numpy as np
from numpy.linalg import norm

class Vector:

    def __init__(self, values: np.ndarray):
        self.values = values
        self.norm: float = norm(values)
        self.size: int = np.size(values)
    
    @staticmethod
    def cosine_similarity(a: "Vector", b: "Vector") -> float:
        return np.dot(a.values, b.values) / (a.norm * b.norm)

    def __len__(self):
        return self.size

    def __eq__(self, value):
        if not isinstance(value, Vector):
            return False
        return self.values == value.values

    def __hash__(self):
        return self.values.__hash__()

    def __repr__(self):
        return self.values.__repr__()



from rb.core.word import Word
from rb.comprehension.utils.distance_strategies.abstract_distance_strategy import DistanceStrategy
from rb.comprehension.utils.graph.cm_node_type import CmNodeType
from rb.comprehension.utils.graph.cm_graph_do import CmGraphDO
from rb.comprehension.utils.graph.cm_node_do import CmNodeDO
from rb.comprehension.utils.graph.cm_edge_do import CmEdgeDO

import numpy as np

from typing import List
Words = List[Word]


class CmWordDistanceIndexer():

    def __init__(self, word_list: Words, strategy: DistanceStrategy):
        self.word_list = word_list
        self.strategy = strategy
        self.index_distances()


    def index_distances(self) -> None:
        n = len(self.word_list)
        self.distances = np.zeros((n, n))

        for i, word1 in enumerate(self.word_list):
            for j, word2 in enumerate(self.word_list):
                if i == j:
                    self.distances[i][j] = self.distances[j][i] = 1
                else:
                    self.distances[i][j] = self.distances[j][i] = self.strategy.get_distance(word1, word2)

        
    def cut_by_avg_plus_stdev(self, minimum_distance: float) -> None:
        threshold = self.get_avg_plus_stdev_threshold(minimum_distance)

        new_word_list = [word for index, word in enumerate(self.word_list)
                            if self.get_max_distance_value_for_word_at_line(index) >= threshold]
        
        self.word_list = new_word_list
        self.index_distances()


    def get_avg_plus_stdev_threshold(self, minimum_distance: float) -> float:
        ok_distances = [self.distances[i][j]
                        for i in range(len(self.word_list) - 1) 
                        for j in range(i + 1, len(self.word_list)) 
                        if self.distances[i][j] > minimum_distance]
        
        if ok_distances:
            avg = np.mean(ok_distances)
            std = np.std(ok_distances)
            return avg - std
        return 0.0


    def get_max_distance_value_for_word_at_line(self, line_number: int) -> float:
        return np.max(self.distances[line_number])

    
    def get_cm_graph(self, node_type: CmNodeType) -> CmGraphDO:
        graph = CmGraphDO([], [])

        node_set = set()
        n = len(self.word_list)
        for i in range(n - 1):
            for j in range(i + 1, n):
                if i != j and self.distances[i][j] > 0:
                    w1 = self.word_list[i]
                    w2 = self.word_list[j]
                    
                    n1 = CmNodeDO(w1, node_type)
                    n1.activate()

                    n2 = CmNodeDO(w2, node_type)
                    n2.activate()

                    edge = CmEdgeDO(n1, n2, self.strategy.get_edge_type(), self.distances[i][j])

                    graph.edge_list.append(edge)
                    node_set.add(n1)
                    node_set.add(n2)

        graph.set_node_list([node for node in node_set if node.word.is_content_word()])

        return graph


from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoCaseEnum(Enum):
    ACC = auto()
    NOM = auto()
    DAT = auto()
    GEN = auto()
    VOC = auto()


class RoPOSFeatureCase(POSFeature):


    _INSTANCE = None

    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.CASE,
                            values=None, pos_supported=None)
        self.values = set([ro_case for _, ro_case in RoCaseEnum.__members__.items()])
        self.pos_supported = set([POS.ADP, POS.PRON, POS.ADJ, POS.DET, POS.NOUN, POS.NUM, POS.PROPN])

    @classmethod
    def get_instance(cls) -> "RoPOSFeatureCase":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeatureCase()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoCaseEnum]:
        pattern = '(Case=[a-zA-Z,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        cases = matches[0].split('=')[1].split(',')
        values = []
        for case in cases:
            if case == 'Dat':
                values.append(RoCaseEnum.DAT)
            elif case == 'Acc':
                values.append(RoCaseEnum.ACC)
            elif case == 'Gen':
                values.append(RoCaseEnum.GEN)
            elif case == 'Nom':
                values.append(RoCaseEnum.NOM)
            elif case == 'Voc':
                values.append(RoCaseEnum.VOC)
        return values




import logging
import logging.config
import os 
import json


class Logger():
    logger = None

    @staticmethod
    def get_logger():
        return Logger.set_up()

    @staticmethod   
    def set_up():
        if Logger.logger is None:
            Logger.setup_logging()
            Logger.logger = logging.getLogger(__name__) 
            Logger.logger.info('Logger set up done.')
            
        for v in logging.Logger.manager.loggerDict.values():
            if isinstance(v, logging.Logger) and not v.name.startswith('rb'):
                v.disabled = True

        return Logger.logger

    @staticmethod
    def setup_logging(
        default_path='logging.json',
        default_level=logging.INFO,
        env_key='LOG_CFG'):
        """Setup logging configuration """
        path = default_path
        if os.path.exists(path):
            with open(path, 'rt') as f:
                config = json.load(f)
            logging.config.dictConfig(config)
        else:
            logging.basicConfig(level=default_level)
        
        

            #if v.name.startswith('rb'):
            # 0   v.disabled = True